"""
Character Helpers
"""

import time
from datetime import timedelta

from django.utils import timezone
from esi.models import Token
from eveuniverse.models import EveType

from ledger.errors import TokenError
from ledger.hooks import get_extension_logger
from ledger.models.characteraudit import (
    CharacterAudit,
    CharacterMiningLedger,
    CharacterWalletJournalEntry,
)
from ledger.models.general import EveEntity
from ledger.providers import esi
from ledger.task_helpers.etag_helpers import NotModifiedError, etag_results

logger = get_extension_logger(__name__)


def get_token(character_id: int, scopes: list) -> "Token":
    """
    Helper method to get a valid token for a specific character with specific scopes.

    Parameters
    ----------
    character_id: `int`
    scopes: `int`

    Returns
    ----------
    `class`: esi.models.Token or False

    """
    token = (
        Token.objects.filter(character_id=character_id)
        .require_scopes(scopes)
        .require_valid()
        .first()
    )
    if token:
        return token
    return False


# pylint: disable=too-many-locals
def update_character_wallet(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(character__character_id=character_id)
    logger.debug(
        "Updating wallet transactions for: %s", audit_char.character.character_name
    )

    req_scopes = ["esi-wallet.read_character_wallet.v1"]

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"

    try:
        journal_items_ob = esi.client.Wallet.get_characters_character_id_wallet_journal(
            character_id=character_id
        )

        journal_items = etag_results(
            journal_items_ob, token, force_refresh=force_refresh
        )

        _st = time.perf_counter()
        _current_journal = CharacterWalletJournalEntry.objects.filter(
            character=audit_char
        ).values_list(
            "entry_id", flat=True
        )  # TODO add time filter
        _current_eve_ids = list(
            EveEntity.objects.all().values_list("eve_id", flat=True)
        )

        _new_names = []

        items = []
        for item in journal_items:
            if item.get("id") not in _current_journal:
                if item.get("second_party_id") not in _current_eve_ids:
                    _new_names.append(item.get("second_party_id"))
                    _current_eve_ids.append(item.get("second_party_id"))
                if item.get("first_party_id") not in _current_eve_ids:
                    _new_names.append(item.get("first_party_id"))
                    _current_eve_ids.append(item.get("first_party_id"))

                # pylint: disable=duplicate-code
                asset_item = CharacterWalletJournalEntry(
                    character=audit_char,
                    amount=item.get("amount"),
                    balance=item.get("balance"),
                    context_id=item.get("context_id"),
                    context_id_type=item.get("context_id_type"),
                    date=item.get("date"),
                    description=item.get("description"),
                    first_party_id=item.get("first_party_id"),
                    entry_id=item.get("id"),
                    reason=item.get("reason"),
                    ref_type=item.get("ref_type"),
                    second_party_id=item.get("second_party_id"),
                    tax=item.get("tax"),
                    tax_receiver_id=item.get("tax_receiver_id"),
                )
                items.append(asset_item)

        created_names = EveEntity.objects.create_bulk_from_esi(_new_names)

        wallet_ballance = esi.client.Wallet.get_characters_character_id_wallet(
            character_id=character_id, token=token.valid_access_token()
        ).result()

        audit_char.balance = wallet_ballance
        audit_char.save()

        if created_names:
            CharacterWalletJournalEntry.objects.bulk_create(items)
        else:
            raise TokenError("ESI Fail")
        logger.debug(
            "TIME: %s update_character_wallet %s",
            time.perf_counter() - _st,
            character_id,
        )
    except NotModifiedError:
        logger.debug("No New wallet data for: %s", audit_char.character.character_name)

    audit_char.last_update_wallet = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return ("Finished wallet transactions for: %s", audit_char.character.character_name)


def update_character_mining(character_id, force_refresh=False):
    audit_char = CharacterAudit.objects.get(character__character_id=character_id)
    logger.debug("Updating Mining for: %s", audit_char.character.character_name)

    req_scopes = ["esi-industry.read_character_mining.v1"]

    token = get_token(character_id, req_scopes)

    if not token:
        return "No Tokens"
    try:
        mining_op = esi.client.Industry.get_characters_character_id_mining(
            character_id=character_id
        )

        ledger = etag_results(mining_op, token, force_refresh=force_refresh)

        _st = time.perf_counter()
        existings_pks = set(
            CharacterMiningLedger.objects.filter(
                character=audit_char, date__gte=timezone.now() - timedelta(days=30)
            ).values_list("id", flat=True)
        )
        type_ids = set()
        new_events = []
        old_events = []
        for event in ledger:

            type_ids.add(event.get("type_id"))
            pk = CharacterMiningLedger.create_primary_key(character_id, event)
            _e = CharacterMiningLedger(
                character=audit_char,
                id=pk,
                date=event.get("date"),
                type_id=event.get("type_id"),
                system_id=event.get("solar_system_id"),
                quantity=event.get("quantity"),
            )
            if pk in existings_pks:
                old_events.append(_e)
            else:
                new_events.append(_e)

        EveType.objects.bulk_get_or_create_esi(ids=list(type_ids))

        if new_events:
            CharacterMiningLedger.objects.bulk_create(new_events, ignore_conflicts=True)

        if old_events:
            CharacterMiningLedger.objects.bulk_update(old_events, fields=["quantity"])

        logger.debug(
            "TIME: %s update_character_mining %s",
            time.perf_counter() - _st,
            character_id,
        )

    except NotModifiedError:
        logger.debug("No New Mining for: %s", audit_char.character.character_name)

    audit_char.last_update_mining = timezone.now()
    audit_char.save()
    audit_char.is_active()

    return f"Finished Mining for: {audit_char.character.character_name}"



from datetime import datetime
from unittest.mock import MagicMock, patch

import _strptime
from bravado.exception import HTTPNotModified

from django.test import TestCase
from esi.errors import TokenError
from esi.models import Token

from ledger.task_helpers.char_helpers import (
    get_token,
    update_character_mining,
    update_character_wallet,
)
from ledger.task_helpers.etag_helpers import NotModifiedError
from ledger.tests.testdata.esi_stub import esi_client_stub
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.task_helpers.char_helpers"


class TestCharacterHelpers(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.mock_token = MagicMock(spec=Token)
        cls.mock_token.character_id = 1001
        cls.mock_token.valid_access_token.return_value = "token"

        cls.journal = [
            {
                "entry_id": 1,
                "amount": 1000,
                "balance": 2000,
                "context_id": 0,
                "context_id_type": "division",
                "date": "2016-10-29T14:00:00Z",
                "description": "Test",
                "first_party_id": 1001,
                "id": 1,
                "reason": "Test",
                "ref_type": "player_division",
                "second_party_id": 1002,
                "tax": 0,
                "tax_receiver_id": 0,
            },
            {
                "entry_id": 2,
                "amount": 1000,
                "balance": 3000,
                "context_id": 0,
                "context_id_type": "division",
                "date": "2016-10-29T14:00:00Z",
                "description": "Test",
                "first_party_id": 1010,
                "id": 2,
                "reason": "Test",
                "ref_type": "player_division",
                "second_party_id": 1020,
                "tax": 0,
                "tax_receiver_id": 0,
            },
            {
                "entry_id": 5,
                "amount": 1000,
                "balance": 4000,
                "context_id": 0,
                "context_id_type": "division",
                "date": "2016-10-29T14:00:00Z",
                "description": "Test",
                "first_party_id": 1035,
                "id": 5,
                "reason": "Test",
                "ref_type": "player_division",
                "second_party_id": 1034,
                "tax": 0,
                "tax_receiver_id": 0,
            },
            {
                "entry_id": 6,
                "amount": 1000,
                "balance": 4000,
                "context_id": 0,
                "context_id_type": "division",
                "date": "2016-10-29T14:00:00Z",
                "description": "Test",
                "first_party_id": 1001,
                "id": 6,
                "reason": "Test",
                "ref_type": "player_division",
                "second_party_id": 1002,
                "tax": 0,
                "tax_receiver_id": 0,
            },
        ]

        cls.mining = [
            {
                "date": datetime.strptime("2024-03-16", "%Y-%m-%d"),
                "quantity": 16751,
                "solar_system_id": 30004783,
                "type_id": 17425,
            },
            {
                "date": datetime.strptime("2024-03-16", "%Y-%m-%d"),
                "quantity": 20000,
                "solar_system_id": 30004783,
                "type_id": 17425,
            },
            {
                "date": datetime.strptime("2024-03-16", "%Y-%m-%d"),
                "quantity": 9810,
                "solar_system_id": 30004783,
                "type_id": 17428,
            },
            {
                "date": datetime.strptime("2024-03-22", "%Y-%m-%d"),
                "quantity": 12550,
                "solar_system_id": 30004783,
                "type_id": 17425,
            },
        ]

        cls.mining2 = [
            {
                "date": datetime.strptime("2024-03-16", "%Y-%m-%d"),
                "quantity": 16751,
                "solar_system_id": 30004783,
                "type_id": 17425,
            },
        ]

    @patch(MODULE_PATH + ".Token.objects.filter")
    def test_get_token(self, mock_token):
        mock_token.return_value.require_scopes.return_value.require_valid.return_value.first.return_value = (
            self.mock_token
        )
        token = get_token(1001, "esi-characters.read_notifications.v1")

        self.assertEqual(token.character_id, 1001)
        self.assertEqual(token.valid_access_token(), "token")

    def test_get_token_no_token(self):
        token = get_token(1001, "esi-characters.read_notifications.v1")

        self.assertFalse(token)

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".CharacterWalletJournalEntry.objects")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".EveEntity.objects.all")
    @patch(MODULE_PATH + ".EveEntity.objects.create_bulk_from_esi")
    @patch(MODULE_PATH + ".CharacterWalletJournalEntry.objects.bulk_create")
    def test_update_character_wallet(
        self,
        mock_charjournal_bulk,
        mock_create_names,
        mock_entity,
        mock_etag,
        mock_esi,
        mock_journal,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = esi_client_stub
        # TODO get Data from Test ESI
        # mock_etag.return_value = mock_esi.client.Wallet.get_characters_character_id_wallet_journal(character_id=1001)
        mock_etag.return_value = self.journal
        mock_journal.filter.return_value.values_list.return_value = [
            1,
            3,
        ]
        mock_entity.return_value.values_list.return_value = [1001]
        mock_create_names.return_value = True
        mock_charjournal_bulk.return_value = None
        # when
        result = update_character_wallet(1001)
        # then
        self.assertEqual(result, ("Finished wallet transactions for: %s", "Gneuten"))

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".CharacterWalletJournalEntry.objects")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".EveEntity.objects.all")
    @patch(MODULE_PATH + ".EveEntity.objects.create_bulk_from_esi")
    @patch(MODULE_PATH + ".CharacterWalletJournalEntry.objects.bulk_create")
    def test_update_character_wallet_token_error(
        self,
        mock_charjournal_bulk,
        mock_create_names,
        mock_entity,
        mock_etag,
        mock_esi,
        mock_journal,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = esi_client_stub
        mock_etag.return_value = self.journal
        mock_journal.filter.return_value.values_list.return_value = [
            1,
            3,
        ]
        mock_entity.return_value.values_list.return_value = [1001, 1002, 1003]
        mock_create_names.return_value = False
        mock_charjournal_bulk.return_value = None
        # when
        # then
        with self.assertRaises(TokenError):
            update_character_wallet(1001)

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".logger")
    @patch(MODULE_PATH + ".etag_results")
    def test_update_character_wallet_not_modified(
        self, mock_etag, mock_logger, mock_get_token
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_etag.side_effect = NotModifiedError
        # when
        update_character_wallet(1001)
        # then
        mock_logger.debug.assert_called_with("No New wallet data for: %s", "Gneuten")

    @patch(MODULE_PATH + ".get_token")
    def test_update_character_wallet_no_token(self, mock_get_token):
        # given
        mock_get_token.return_value = None
        # when
        result = update_character_wallet(1001)
        # then
        self.assertEqual(result, ("No Tokens"))

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CharacterMiningLedger.objects")
    @patch(MODULE_PATH + ".EveType.objects.bulk_get_or_create_esi")
    def test_update_character_mining(
        self, _, mock_existing, mock_etag, mock_esi, mock_get_token
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = esi_client_stub
        mock_etag.return_value = self.mining
        mock_existing.filter.return_value.values_list.return_value = [
            "20240316-17425-1001-30004783"
        ]
        # when
        result = update_character_mining(1001)
        # then
        self.assertEqual(result, ("Finished Mining for: Gneuten"))

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CharacterMiningLedger.objects")
    @patch(MODULE_PATH + ".EveType.objects.bulk_get_or_create_esi")
    def test_update_character_mining_non_existing(
        self, _, mock_existing, mock_etag, mock_esi, mock_get_token
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = esi_client_stub
        mock_etag.return_value = self.mining
        mock_existing.filter.return_value.values_list.return_value = []
        # when
        result = update_character_mining(1001)
        # then
        self.assertEqual(result, ("Finished Mining for: Gneuten"))

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".CharacterMiningLedger.objects")
    @patch(MODULE_PATH + ".EveType.objects.bulk_get_or_create_esi")
    def test_update_character_mining_2(
        self, _, mock_existing, mock_etag, mock_esi, mock_get_token
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_esi.client = esi_client_stub
        mock_etag.return_value = self.mining2
        mock_existing.filter.return_value.values_list.return_value = [
            "20240316-17425-1001-30004783"
        ]
        # when
        result = update_character_mining(1001)
        # then
        self.assertEqual(result, ("Finished Mining for: Gneuten"))

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".etag_results")
    def test_update_character_mining_no_token(self, _, mock_esi, mock_get_token):
        # given
        mock_get_token.return_value = None
        # when
        result = update_character_mining(1001)
        # then
        self.assertEqual(result, ("No Tokens"))

    @patch(MODULE_PATH + ".get_token")
    @patch(MODULE_PATH + ".logger")
    @patch(MODULE_PATH + ".etag_results")
    @patch(MODULE_PATH + ".esi.client.Industry.get_characters_character_id_mining")
    def test_update_character_mining_not_modified(
        self,
        mock_get_characters_character_id_mining,
        mock_etag,
        mock_logger,
        mock_get_token,
    ):
        # given
        mock_get_token.return_value = self.mock_token
        mock_etag.side_effect = NotModifiedError
        mock_get_characters_character_id_mining.side_effect = HTTPNotModified
        # when
        update_character_mining(1001)
        # then
        mock_logger.debug.assert_called_with("No New Mining for: %s", "Gneuten")



from django.core.exceptions import ObjectDoesNotExist
from django.db import models
from django.db.models import Case, DecimalField, OuterRef, Q, Subquery, Sum, Value, When
from django.db.models.functions import Coalesce

from allianceauth.eveonline.models import EveCharacter

from ledger.hooks import get_corp_models_and_string, get_extension_logger
from ledger.view_helpers.core import events_filter

logger = get_extension_logger(__name__)


class CorpWalletQuerySet(models.QuerySet):
    def _get_linked_chars(self, corporations: list) -> dict:
        linked_chars = EveCharacter.objects.filter(corporation_id__in=corporations)
        linked_chars |= EveCharacter.objects.filter(
            character_ownership__user__profile__main_character__corporation_id__in=corporations
        )
        linked_chars = linked_chars.select_related(
            "character_ownership", "character_ownership__user__profile__main_character"
        ).prefetch_related("character_ownership__user__character_ownerships__character")

        corpmember = get_corp_models_and_string()
        char_to_main = {}
        main_to_alts = {}
        for char in linked_chars:
            try:
                if char.corporation_id in corporations:
                    main_char = char.character_ownership.user.profile.main_character
                    main_char_id = main_char.character_id
                    char_to_main[char.character_id] = main_char_id
                if main_char_id not in main_to_alts:
                    main_to_alts[main_char_id] = []
                main_to_alts[main_char_id].append(char.character_id)
            except ObjectDoesNotExist:
                continue
            except AttributeError:
                continue

        corpmember = (
            corpmember.objects.filter(corpstats__corp__corporation_id__in=corporations)
            .values_list("character_id", flat=True)
            .exclude(character_id__in=char_to_main)
        )

        for character_id in corpmember:
            char_to_main[character_id] = character_id
        return char_to_main, main_to_alts

    def annotate_bounty(self, second_party_ids: list) -> models.QuerySet:
        return self.annotate(
            total_bounty=Coalesce(
                Sum(
                    "amount",
                    filter=(
                        Q(ref_type="bounty_prizes")
                        & Q(second_party_id__in=second_party_ids)
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_ess(self, second_party_ids: list) -> models.QuerySet:
        qs = events_filter(self)
        return qs.annotate(
            total_ess=Coalesce(
                Sum(
                    "amount",
                    filter=(
                        Q(ref_type="ess_escrow_transfer")
                        & Q(second_party_id__in=second_party_ids)
                    ),
                ),
                Value(0),
                output_field=DecimalField(),
            )
        )

    def annotate_ledger(self, corporations: list) -> models.QuerySet:
        char_to_main, main_to_alts = self._get_linked_chars(corporations)

        # Create a subquery to map second_party_id to main_character_id
        subquery = Case(
            *[
                When(second_party_id=char_id, then=main_id)
                for char_id, main_id in char_to_main.items()
            ],
            output_field=models.IntegerField(),
        )

        return (
            self.filter(second_party_id__in=char_to_main.keys())
            .annotate(
                main_character_id=subquery,
                main_character_name=Subquery(
                    EveCharacter.objects.filter(
                        character_id=OuterRef("main_character_id")
                    ).values("character_name")[:1]
                ),
                alts=Value(
                    [
                        alt_id
                        for main_id, alts in main_to_alts.items()
                        for alt_id in alts
                        if main_id == OuterRef("main_character_id")
                    ],
                    output_field=models.JSONField(),
                ),
            )
            .values(
                "main_character_id", "main_character_name", "alts"
            )  # Group by main_character_id
            .annotate(
                total_bounty=Coalesce(
                    Sum("amount", filter=Q(ref_type="bounty_prizes")),
                    Value(0),
                    output_field=DecimalField(),
                ),
                total_ess=Coalesce(
                    Sum("amount", filter=Q(ref_type="ess_escrow_transfer")),
                    Value(0),
                    output_field=DecimalField(),
                ),
            )
            .filter(Q(total_bounty__gt=0) | Q(total_ess__gt=0))
        )


class CorpWalletManagerBase(models.Manager):
    pass


CorpWalletManager = CorpWalletManagerBase.from_queryset(CorpWalletQuerySet)



"""
Character Audit
"""

from datetime import timedelta

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.shortcuts import redirect
from django.utils import timezone
from esi.decorators import token_required

from allianceauth.eveonline.models import EveCharacter

from ledger.hooks import get_extension_logger
from ledger.models.characteraudit import CharacterAudit
from ledger.tasks import update_character

logger = get_extension_logger(__name__)


@login_required
@token_required(scopes=CharacterAudit.get_esi_scopes())
def add_char(request, token):
    CharacterAudit.objects.update_or_create(
        character=EveCharacter.objects.get_character_by_id(token.character_id)
    )
    update_character.apply_async(
        args=[token.character_id], kwargs={"force_refresh": True}, priority=6
    )
    msg = "Char successfully added/updated to Ledger"
    messages.info(request, msg)
    return redirect("ledger:ledger_index")


@login_required
@permission_required(["ledger.admin_access"])
def fetch_memberaudit(request):
    """Add All Chars from Memberaudit to CharacterAudit"""
    try:
        # pylint: disable=import-outside-toplevel
        from memberaudit.models import Character

        chars = Character.objects.filter(is_disabled=0)
        ledger_chars_ids = CharacterAudit.objects.values_list(
            "character__character_id", flat=True
        )
        count = 0

        # Zeitlimit für den letzten Wallet-Update (2 Stunden)
        time_limit = timezone.now() - timedelta(hours=2)

        for char in chars:
            try:
                if char.eve_character.character_id in ledger_chars_ids:
                    continue
                char_audit, _ = CharacterAudit.objects.get_or_create(
                    character=char.eve_character,
                    id=char.id,
                )
                if char_audit:
                    if (
                        not char_audit.last_update_wallet
                        or char_audit.last_update_wallet < time_limit
                    ):
                        update_character.apply_async(
                            args=[char.eve_character.character_id],
                            kwargs={"force_refresh": True},
                            priority=6,
                        )
                        count += 1
                    else:
                        continue
            except EveCharacter.DoesNotExist:
                logger.debug("Character %s not found", char)
                continue
            except AttributeError:
                logger.debug("Attribute Error for %s", char)
                continue
        msg = (
            f"{count} Char(s) successfully added/updated to Ledger"
            if count
            else "No Updates initialized."
        )
        messages.info(request, msg)
    except ImportError:
        msg = (
            "The 'memberaudit' app is not installed. Please make sure it is installed."
        )
        messages.error(request, msg)

    except Exception as e:  # pylint: disable=broad-exception-caught
        logger.error(e, exc_info=True)
        msg = "An error occurred: Please inform your Admin."
        messages.error(request, msg)

    return redirect("ledger:ledger_index")



from django.test import RequestFactory, TestCase

from allianceauth.eveonline.models import EveCorporationInfo
from app_utils.testing import create_user_from_evecharacter

from ledger.models.corporationaudit import (
    CorporationAudit,
    CorporationWalletJournalEntry,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.models.general"


class TestCorporationAuditModel(TestCase):
    def setUp(self):
        load_allianceauth()
        self.audit = CorporationAudit(
            corporation=EveCorporationInfo.objects.get(corporation_id=2001)
        )

    def test_str(self):
        self.assertEqual(str(self.audit), "Hell RiderZ's Corporation Data")


class TestCorporationWalletJournal(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        cls.journal = CorporationWalletJournalEntry.objects.get(entry_id=1)

    def test_str(self):
        self.assertEqual(
            str(self.journal),
            "Corporation Wallet Journal: CONCORD 'ess_escrow_transfer' Gneuten: 100000.00 isk",
        )

    def test_get_visible(self):
        self.factory = RequestFactory()
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.corp_audit_admin_manager",
            ],
        )
        request = self.factory.get("/")
        request.user = self.user

        query = CorporationWalletJournalEntry.get_visible(request.user)

        excepted_corporation = CorporationWalletJournalEntry.objects.all()

        self.assertEqual(list(query), list(excepted_corporation))



from unittest.mock import MagicMock, patch

from django.test import TestCase

from ledger.task_helpers.core_helpers import (
    AlreadyQueued,
    enqueue_next_task,
    no_fail_chain,
)

MODULE_PATH = "ledger.task_helpers.core_helpers"


class CoreHelpersTest(TestCase):
    @patch(MODULE_PATH + ".enqueue_next_task")
    def test_no_fail_chain(self, mock_enqueue):
        @no_fail_chain
        def test_func(*args, **kwargs):
            raise Exception("Test exception")

        with self.assertRaises(Exception):
            test_func(chain=[])

        mock_enqueue.assert_called_once()

    @patch(MODULE_PATH + ".signature")
    def test_enqueue_next_task(self, mock_signature):
        # given
        mock_task1 = MagicMock()
        mock_task2 = MagicMock()
        chain = [mock_task1, mock_task2]

        mock_sig = MagicMock()
        mock_sig.kwargs = {}  # Set kwargs to an empty dictionary
        mock_signature.return_value = mock_sig
        # when
        enqueue_next_task(chain)
        # then
        mock_sig.apply_async.assert_called_once_with(priority=9)
        self.assertEqual(mock_sig.kwargs["chain"], [mock_task2])

    @patch(MODULE_PATH + ".signature")
    @patch(MODULE_PATH + ".logger")
    def test_enqueue_next_task_alreadyqueued(self, mock_logger, mock_signature):
        # given
        mock_task1 = MagicMock()
        mock_task2 = MagicMock()
        chain = [mock_task1, mock_task2]

        mock_sig = MagicMock()
        mock_sig.kwargs = {}  # Set kwargs to an empty dictionary
        mock_sig.apply_async.side_effect = AlreadyQueued(
            countdown=10
        )  # apply_async raises AlreadyQueued
        mock_signature.return_value = mock_sig
        # when
        enqueue_next_task(chain)
        # then
        mock_sig.apply_async.assert_called()
        self.assertEqual(chain, [])
        mock_logger.debug.assert_called_with(
            "Skipping task as its already queued %s", mock_sig
        )



from http import HTTPStatus

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.test import RequestFactory, TestCase
from django.urls import reverse

from app_utils.testdata_factories import UserMainFactory

from ledger.models.general import General
from ledger.views.character.character_ledger import character_admin, character_ledger
from ledger.views.corporation.corporation_ledger import (
    corporation_admin,
    corporation_ledger,
)
from ledger.views.pve import ledger_index


class TestViews(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.user = UserMainFactory(
            permissions=[
                "ledger.basic_access",
            ]
        )

    def test_view(self):
        request = self.factory.get(reverse("ledger:index"))
        request.user = self.user
        response = ledger_index(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_char_ledger_view(self):
        request = self.factory.get(
            reverse("ledger:character_ledger", kwargs={"character_pk": 0})
        )
        request.user = self.user
        response = character_ledger(request, character_pk=0)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_char_ledger_admin_view(self):
        request = self.factory.get(reverse("ledger:character_admin"))
        request.user = self.user
        response = character_admin(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_corp_ledger_view(self):
        request = self.factory.get(
            reverse("ledger:corporation_ledger", kwargs={"corporation_pk": 0})
        )
        request.user = self.user
        response = corporation_ledger(request, corporation_pk=0)
        self.assertEqual(response.status_code, HTTPStatus.OK)

    def test_corp_ledger_admin_view(self):
        request = self.factory.get(reverse("ledger:corporation_admin"))
        request.user = self.user
        response = corporation_admin(request)
        self.assertEqual(response.status_code, HTTPStatus.OK)



from decimal import Decimal
from unittest.mock import MagicMock, patch

from django.test import TestCase
from esi.models import Token

from allianceauth.corputils.models import CorpMember, CorpStats
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ledger.managers.corpjournal_manager import CorpWalletManager, CorpWalletManagerBase
from ledger.models import CorporationWalletJournalEntry
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.managers.corpjournal_manager"


class CharManagerQuerySetTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()
        cls.manager = CorpWalletManager()
        cls.corp = EveCorporationInfo.objects.get(corporation_id=2001)

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )

    def test_annotate_bounty(self):
        character_ids = [1, 2, 3]

        qs = CorporationWalletJournalEntry.objects.annotate_bounty(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_bounty", qs.query.annotations)

    def test_annotate_ess(self):
        character_ids = [1, 2, 3]

        qs = CorporationWalletJournalEntry.objects.annotate_ess(character_ids)
        self.assertIsNotNone(qs)
        self.assertIn("total_ess", qs.query.annotations)

    def test_annotate_ledger(self):
        add_character_to_user(self.user, EveCharacter.objects.get(character_id=1002))
        add_character_to_user(self.user, EveCharacter.objects.get(character_id=1004))
        add_character_to_user(self.user, EveCharacter.objects.get(character_id=1003))
        add_character_to_user(self.user, EveCharacter.objects.get(character_id=1006))

        corp_stats = CorpStats.objects.create(
            token=Token.objects.filter(user=self.user).first(),
            corp=self.corp,
        )
        CorpMember.objects.create(
            character_id=9999, character_name="Test9999", corpstats=corp_stats
        )

        self.client.force_login(self.user)
        result = CorporationWalletJournalEntry.objects.all().annotate_ledger([2001])

        expected_result = [
            {
                "main_character_id": 1001,
                "main_character_name": "Gneuten",
                "alts": [],
                "total_bounty": Decimal("400000.00"),
                "total_ess": Decimal("400000.00"),
            }
        ]

        self.assertEqual(list(result), expected_result)

    def test_annotate_ledger_with_attribute_error(self):
        auth_character = EveCharacter.objects.get(character_id=1019)
        user = AuthUtils.create_user(auth_character.character_name.replace(" ", "_"))
        _ = add_character_to_user(user, auth_character, is_main=False, scopes=None)

        self.client.force_login(user)

        result = CorporationWalletJournalEntry.objects.all().annotate_ledger([2001])

        expected_result = [
            {
                "main_character_id": 1001,
                "main_character_name": "Gneuten",
                "alts": [],
                "total_bounty": Decimal("400000.00"),
                "total_ess": Decimal("400000.00"),
            }
        ]

        self.assertEqual(list(result), expected_result)

    def test_annotate_ledger_no_data(self):
        result = CorporationWalletJournalEntry.objects.all().annotate_ledger([9999])

        self.assertEqual(list(result), [])

    def test_annotate_ledger_no_data_no_corporations(self):
        result = CorporationWalletJournalEntry.objects.all().annotate_ledger([])

        self.assertEqual(list(result), [])



# Generated by Django 4.2.11 on 2024-05-16 12:14

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("eveuniverse", "0010_alter_eveindustryactivityduration_eve_type_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app, Ledger."),
                    ("moderator_access", "Has access to moderation tools"),
                    ("admin_access", "Has access to all Administration tools"),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CharacterAudit",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("active", models.BooleanField(default=False)),
                (
                    "last_update_wallet",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "last_update_mining",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "balance",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=20, null=True
                    ),
                ),
                (
                    "character",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_character",
                        to="eveonline.evecharacter",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("char_audit_admin_access", "Has access to all characters"),
                ),
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CorporationAudit",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "last_update_wallet",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "corporation",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_corporationaudit",
                        to="eveonline.evecorporationinfo",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("corp_audit_admin_access", "Has access to all Corporations"),
                ),
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CorporationWalletDivision",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(default=None, max_length=100, null=True)),
                ("balance", models.DecimalField(decimal_places=2, max_digits=20)),
                ("division", models.IntegerField()),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_corporation_division",
                        to="ledger.corporationaudit",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CorpSteuer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("character_id", models.PositiveIntegerField(null=True)),
                ("character_name", models.CharField(max_length=100, null=True)),
                ("status", models.CharField(blank=True, max_length=50, null=True)),
                ("date", models.DateField(blank=True, null=True)),
            ],
            options={
                "permissions": (
                    ("steuer_admin_access", "Has access to Tax Administration"),
                ),
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Events",
            fields=[
                ("id", models.BigAutoField(primary_key=True, serialize=False)),
                ("title", models.CharField(max_length=100, null=True)),
                ("date_start", models.DateTimeField(blank=True, null=True)),
                ("date_end", models.DateTimeField(blank=True, null=True)),
                ("description", models.TextField(blank=True, null=True)),
                ("char_ledger", models.BooleanField(default=False, null=True)),
                ("location", models.CharField(blank=True, max_length=100, null=True)),
            ],
            options={
                "verbose_name": "event",
                "verbose_name_plural": "events",
                "permissions": (("event_admin_access", "Can access Events Tools"),),
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="EveEntity",
            fields=[
                (
                    "eve_id",
                    models.IntegerField(
                        primary_key=True,
                        serialize=False,
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="id",
                    ),
                ),
                (
                    "category",
                    models.CharField(
                        choices=[
                            ("alliance", "Alliance"),
                            ("corporation", "Corporation"),
                            ("character", "Character"),
                        ],
                        max_length=32,
                        verbose_name="category",
                    ),
                ),
                ("name", models.CharField(max_length=254, verbose_name="name")),
                ("last_update", models.DateTimeField(auto_now=True)),
                (
                    "alliance",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="alli",
                        to="ledger.eveentity",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="corp",
                        to="ledger.eveentity",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CharacterMiningLedger",
            fields=[
                (
                    "id",
                    models.CharField(max_length=50, primary_key=True, serialize=False),
                ),
                ("date", models.DateField()),
                ("quantity", models.IntegerField()),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_character",
                        to="ledger.characteraudit",
                    ),
                ),
                (
                    "system",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_evesolarsystem",
                        to="eveuniverse.evesolarsystem",
                    ),
                ),
                (
                    "type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="ledger_evetype",
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="CorporationWalletJournalEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=20, null=True
                    ),
                ),
                (
                    "balance",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=20, null=True
                    ),
                ),
                ("context_id", models.BigIntegerField(default=None, null=True)),
                (
                    "context_id_type",
                    models.CharField(
                        choices=[
                            ("structure_id", "Structure Id"),
                            ("station_id", "Station Id"),
                            ("market_transaction_id", "Market Transaction Id"),
                            ("character_id", "Character Id"),
                            ("corporation_id", "Corporation Id"),
                            ("alliance_id", "Alliance Id"),
                            ("eve_system", "Eve System"),
                            ("industry_job_id", "Industry Job Id"),
                            ("contract_id", "Contract Id"),
                            ("planet_id", "Planet Id"),
                            ("system_id", "System Id"),
                            ("type_id", "Type Id"),
                        ],
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                ("date", models.DateTimeField()),
                ("description", models.CharField(max_length=500)),
                ("entry_id", models.BigIntegerField()),
                ("reason", models.CharField(default=None, max_length=500, null=True)),
                ("ref_type", models.CharField(max_length=72)),
                (
                    "tax",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=20, null=True
                    ),
                ),
                ("tax_receiver_id", models.IntegerField(default=None, null=True)),
                (
                    "division",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="ledger.corporationwalletdivision",
                    ),
                ),
                (
                    "first_party",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="ledger.eveentity",
                    ),
                ),
                (
                    "second_party",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="ledger.eveentity",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "default_permissions": (),
                "indexes": [
                    models.Index(fields=["date"], name="ledger_corp_date_c08a55_idx"),
                    models.Index(
                        fields=["amount"], name="ledger_corp_amount_e482ac_idx"
                    ),
                    models.Index(
                        fields=["entry_id"], name="ledger_corp_entry_i_e36818_idx"
                    ),
                    models.Index(
                        fields=["ref_type"], name="ledger_corp_ref_typ_280f5b_idx"
                    ),
                    models.Index(
                        fields=["first_party"], name="ledger_corp_first_p_114662_idx"
                    ),
                    models.Index(
                        fields=["second_party"], name="ledger_corp_second__9a820c_idx"
                    ),
                ],
            },
        ),
        migrations.CreateModel(
            name="CharacterWalletJournalEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=20, null=True
                    ),
                ),
                (
                    "balance",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=20, null=True
                    ),
                ),
                ("context_id", models.BigIntegerField(default=None, null=True)),
                (
                    "context_id_type",
                    models.CharField(
                        choices=[
                            ("structure_id", "Structure Id"),
                            ("station_id", "Station Id"),
                            ("market_transaction_id", "Market Transaction Id"),
                            ("character_id", "Character Id"),
                            ("corporation_id", "Corporation Id"),
                            ("alliance_id", "Alliance Id"),
                            ("eve_system", "Eve System"),
                            ("industry_job_id", "Industry Job Id"),
                            ("contract_id", "Contract Id"),
                            ("planet_id", "Planet Id"),
                            ("system_id", "System Id"),
                            ("type_id", "Type Id"),
                        ],
                        default=None,
                        max_length=30,
                        null=True,
                    ),
                ),
                ("date", models.DateTimeField()),
                ("description", models.CharField(max_length=500)),
                ("entry_id", models.BigIntegerField()),
                ("reason", models.CharField(default=None, max_length=500, null=True)),
                ("ref_type", models.CharField(max_length=72)),
                (
                    "tax",
                    models.DecimalField(
                        decimal_places=2, default=None, max_digits=20, null=True
                    ),
                ),
                ("tax_receiver_id", models.IntegerField(default=None, null=True)),
                (
                    "character",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="ledger.characteraudit",
                    ),
                ),
                (
                    "first_party",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="ledger.eveentity",
                    ),
                ),
                (
                    "second_party",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="ledger.eveentity",
                    ),
                ),
            ],
            options={
                "abstract": False,
                "default_permissions": (),
                "indexes": [
                    models.Index(fields=["date"], name="ledger_char_date_e50f13_idx"),
                    models.Index(
                        fields=["amount"], name="ledger_char_amount_2c2575_idx"
                    ),
                    models.Index(
                        fields=["entry_id"], name="ledger_char_entry_i_8f5952_idx"
                    ),
                    models.Index(
                        fields=["ref_type"], name="ledger_char_ref_typ_f611b2_idx"
                    ),
                    models.Index(
                        fields=["first_party"], name="ledger_char_first_p_c0bd40_idx"
                    ),
                    models.Index(
                        fields=["second_party"], name="ledger_char_second__aa3ded_idx"
                    ),
                ],
            },
        ),
    ]



import calendar
from dataclasses import asdict, dataclass, field
from datetime import datetime

from django.db.models import Q
from django.utils import timezone

from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


@dataclass
class LedgerDataCore:
    total_bounty: int = 0
    total_ess_payout: int = 0
    total_mining: int = 0
    total_miscellaneous: int = 0
    total_isk: int = 0


@dataclass
class LedgerData(LedgerDataCore):
    total_cost: int = 0
    total_production_cost: int = 0
    total_market_cost: int = 0


class LedgerModels:
    """LedgerModels class to store the models."""

    def __init__(
        self, character_journal=None, corporation_journal=None, mining_journal=None
    ):
        self.char_journal = character_journal
        self.corp_journal = corporation_journal
        self.mining_journal = mining_journal


@dataclass
class LedgerDate:
    year: int
    month: int
    monthly: bool = field(init=False)
    current_date: datetime = None
    range_data: int = field(init=False)
    day_checks: list = field(init=False)

    def calculate_days(self):
        _, num_days = calendar.monthrange(self.year, self.month)
        return num_days

    def __post_init__(self):
        self.current_date = timezone.now()
        self.monthly = self.month == 0
        self.range_data = 12 if self.monthly else self.calculate_days()
        self.day_checks = list(range(1, self.range_data + 1))


@dataclass
class LedgerTotal:
    total_amount: int = 0
    total_amount_ess: int = 0
    total_amount_all: int = 0
    total_amount_mining: int = 0
    total_amount_others: int = 0
    total_amount_costs: int = 0

    def to_dict(self):
        return asdict(self)

    def get_data(self, totals: dict):
        self.total_amount += totals.get("total_amount", 0)
        self.total_amount_ess += totals.get("total_amount_ess", 0)
        self.total_amount_all += totals.get("total_amount_all", 0)
        self.total_amount_mining += totals.get("total_amount_mining", 0)
        self.total_amount_others += totals.get("total_amount_others", 0)
        self.total_amount_costs += totals.get("total_amount_costs", 0)


@dataclass
class LedgerFilterCore:
    char_id: list

    filter_first_party: Q = field(init=False)
    filter_second_party: Q = field(init=False)
    filter_partys: Q = field(init=False)
    filter_total: Q = field(init=False)

    def __post_init__(self):
        self.filter_first_party = Q(first_party_id__in=self.char_id)
        self.filter_second_party = Q(second_party_id__in=self.char_id)
        self.filter_partys = self.filter_first_party | self.filter_second_party
        self.filter_total = self.filter_partys & Q(amount__gt=0)


@dataclass
class LedgerFilterPvE(LedgerFilterCore):

    def __init__(self, char_id):
        super().__init__(char_id)

        self.filter_bounty = self.filter_second_party & Q(
            ref_type="bounty_prizes", amount__gt=0
        )
        self.filter_ess = self.filter_second_party & Q(
            ref_type="ess_escrow_transfer", amount__gt=0
        )
        self.filter_mining = Q(character__character__character_id__in=self.char_id)

        self.filter_all_pve = self.filter_bounty | self.filter_ess | self.filter_mining

    # TODO add mining filter with aggregate
    def get_all_pve_filters(self):
        return {
            "bounty": self.filter_bounty,
            # "ess": self.filter_ess,
        }

    def get_corp_filters(self):
        return {
            "bounty": self.filter_bounty,
            "ess": self.filter_ess,
        }


# pylint: disable=too-many-instance-attributes
@dataclass
class LedgerFilterCost(LedgerFilterPvE):
    """LedgerFilter class to store the filter data."""

    def __init__(self, char_id):
        super().__init__(char_id)
        self.filter_contract_cost = self.filter_partys & Q(
            ref_type__in=[
                "contract_price_payment_corp",
                "contract_reward",
                "contract_price",
                "contract_collateral",
                "contract_reward_deposited",
            ],
            amount__lt=0,
        )
        self.filter_market_cost = self.filter_partys & Q(
            ref_type__in=[
                "market_escrow",
                "transaction_tax",
                "market_provider_tax",
                "brokers_fee",
            ],
            amount__lt=0,
        )
        self.filter_assets_cost = self.filter_partys & Q(
            ref_type__in=[
                "asset_safety_recovery_tax",
            ],
            amount__lt=0,
        )
        self.filter_traveling_cost = self.filter_partys & Q(
            ref_type__in=[
                "structure_gate_jump",
                "jump_clone_activation_fee",
            ],
            amount__lt=0,
        )
        self.filter_production_cost = self.filter_partys & Q(
            ref_type__in=[
                "industry_job_tax",
                "manufacturing",
                "researching_time_productivity",
                "researching_material_productivity",
                "copying",
                "reprocessing_tax",
                "reaction",
            ],
            amount__lt=0,
        )

        self.filter_skill_cost = self.filter_partys & Q(
            ref_type__in=[
                "skill_purchase",
            ],
            amount__lt=0,
        )

        self.filter_insurance_cost = self.filter_partys & Q(
            ref_type__in=[
                "insurance",
            ],
            amount__lt=0,
        )

        self.filter_planetary_cost = self.filter_partys & Q(
            ref_type__in=[
                "planetary_import_tax",
                "planetary_export_tax",
                "planetary_construction",
            ],
            amount__lt=0,
        )

        self.loyality_points_cost = self.filter_partys & Q(
            ref_type="lp_store", amount__lt=0
        )

        self.filter_costs = self.filter_partys & Q(amount__lt=0)

        self.filter_all_costs = (
            self.filter_contract_cost
            | self.filter_market_cost
            | self.filter_assets_cost
            | self.filter_traveling_cost
            | self.filter_production_cost
            | self.filter_skill_cost
            | self.filter_insurance_cost
            | self.filter_planetary_cost
            | self.loyality_points_cost
        )

    def get_all_costs_filters(self):
        return {
            "market_cost": self.filter_market_cost,
            "production_cost": self.filter_production_cost,
            "contract_cost": self.filter_contract_cost,
            "traveling_cost": self.filter_traveling_cost,
            "asset_cost": self.filter_assets_cost,
            "skill_cost": self.filter_skill_cost,
            "insurance_cost": self.filter_insurance_cost,
            "planetary_cost": self.filter_planetary_cost,
            "loyality_point_cost": self.loyality_points_cost,
        }


@dataclass
class LedgerFilterTrading(LedgerFilterCost):
    """LedgerFilter class to store the filter data."""

    def __init__(self, char_id):
        super().__init__(char_id)
        self.filter_market = self.filter_partys & Q(
            ref_type="market_transaction", amount__gt=0
        )

        self.filter_contract = self.filter_partys & Q(
            ref_type__in=[
                "contract_price_payment_corp",
                "contract_reward",
                "contract_price",
            ],
            amount__gt=0,
        )

        self.filter_donation = self.filter_partys & Q(
            ref_type="player_donation", amount__gt=0
        )

        self.filter_insurance = self.filter_second_party & Q(
            ref_type__in=[
                "insurance",
            ],
            amount__gt=0,
        )

        self.filter_all_misc = (
            self.filter_market | self.filter_contract | self.filter_insurance
        )

    def get_all_misc_filters(self, chars_list):
        return {
            "transaction": self.filter_market,
            "contract": self.filter_contract,
            "donation": self.filter_donation & ~Q(first_party_id__in=chars_list),
            "insurance": self.filter_insurance,
        }


@dataclass
class LedgerFilterMission(LedgerFilterTrading):
    """LedgerFilterMission class to store the filter data."""

    def __init__(self, char_id):
        super().__init__(char_id)
        self.filter_mission = self.filter_partys & Q(
            ref_type__in=[
                "agent_mission_reward",
                "agent_mission_time_bonus_reward",
            ],
            amount__gt=0,
        )

        self.filter_all_missions = self.filter_mission

    def get_all_mission_filters(self):
        return {
            "mission": self.filter_mission,
        }


@dataclass
class LedgerFilter(LedgerFilterMission):
    """LedgerFilterAll class to store all filter data."""

    def __init__(self, char_id):
        super().__init__(char_id)
        self.char_id = char_id

    def get_all_filters(self, chars_list):
        filters = {}
        filters.update(self.get_all_pve_filters())
        filters.update(self.get_all_costs_filters())
        filters.update(self.get_all_misc_filters(chars_list))
        filters.update(self.get_all_mission_filters())
        return filters



"""App Tasks"""

# Standard Library
import datetime

# Third Party
# pylint: disable=no-name-in-module
from celery import shared_task

from django.db import IntegrityError
from django.utils import timezone
from esi.errors import TokenExpiredError
from esi.models import Token

from allianceauth.authentication.models import EveCharacter
from allianceauth.services.tasks import QueueOnce

from ledger.decorators import when_esi_is_available
from ledger.hooks import get_extension_logger
from ledger.models.characteraudit import CharacterAudit
from ledger.models.corporationaudit import CorporationAudit
from ledger.task_helpers.char_helpers import (
    update_character_mining,
    update_character_wallet,
)
from ledger.task_helpers.core_helpers import enqueue_next_task, no_fail_chain
from ledger.task_helpers.corp_helpers import update_corp_wallet_division

logger = get_extension_logger(__name__)


# Member Audit Adaptation
# pylint: disable=unused-argument
@shared_task(bind=True, base=QueueOnce)
@when_esi_is_available
def create_member_audit(self, runs: int = 0):
    # pylint: disable=import-outside-toplevel
    from memberaudit.models import Character

    chars_ids = CharacterAudit.objects.all().values_list(
        "character__character_id", flat=True
    )
    member_ids = Character.objects.all().values_list(
        "eve_character__character_id", flat=True
    )

    new_ids = set(member_ids) - set(chars_ids)

    for character_id in new_ids:
        try:
            CharacterAudit.objects.update_or_create(
                character=EveCharacter.objects.get_character_by_id(character_id)
            )
            runs = runs + 1
        except IntegrityError:
            continue
    logger.debug("Created %s missing Member Audit Characters", runs)
    return True


# Character Audit - Tasks
# pylint: disable=unused-argument
@shared_task(bind=True, base=QueueOnce)
@when_esi_is_available
def create_missing_character(self, chars_list: list, runs: int = 0):
    for character_id in chars_list:
        try:
            EveCharacter.objects.create_character(
                character_id=character_id,
            )
            runs = runs + 1
        except IntegrityError:
            continue
    logger.debug("Created %s missing Characters", runs)
    return True


@shared_task
@when_esi_is_available
def update_all_characters(runs: int = 0):
    characters = CharacterAudit.objects.select_related("character").all()
    for char in characters:
        update_character.apply_async(args=[char.character.character_id])
        runs = runs + 1
    logger.info("Queued %s Char Audit Updates", runs)


@shared_task(bind=True, base=QueueOnce)
def update_character(
    self, char_id, force_refresh=False
):  # pylint: disable=unused-argument
    character = (
        CharacterAudit.objects.select_related("character")
        .filter(character__character_id=char_id)
        .first()
    )
    if character is None:
        token = Token.get_token(char_id, CharacterAudit.get_esi_scopes())
        if token:
            try:
                if token.valid_access_token():
                    character, _ = CharacterAudit.objects.update_or_create(
                        character=EveCharacter.objects.get_character_by_id(
                            token.character_id
                        )
                    )
                else:
                    return False
            except TokenExpiredError:
                return False
        else:
            logger.info("No Tokens for %s", char_id)
            return False

    logger.debug(
        "Processing Audit Updates for %s", format(character.character.character_name)
    )

    skip_date = timezone.now() - datetime.timedelta(hours=1)
    que = []
    mindt = timezone.now() - datetime.timedelta(days=90)

    if (character.last_update_mining or mindt) <= skip_date or force_refresh:
        que.append(
            update_char_mining_ledger.si(
                character.character.character_id, force_refresh=force_refresh
            )
        )

    if (character.last_update_wallet or mindt) <= skip_date or force_refresh:
        que.append(
            update_char_wallet.si(
                character.character.character_id, force_refresh=force_refresh
            )
        )

    enqueue_next_task(que)

    logger.debug(
        "Queued %s Audit Updates for %s",
        len(que) + 1,
        character.character.character_name,
    )

    return True


# Corporation Audit - Tasks


@shared_task
@when_esi_is_available
def update_all_corps(runs: int = 0):
    corps = CorporationAudit.objects.select_related("corporation").all()
    for corp in corps:
        update_corp.apply_async(args=[corp.corporation.corporation_id])
        runs = runs + 1
    logger.info("Queued %s Corp Audit Updates", runs)


@shared_task(bind=True, base=QueueOnce)
def update_corp(self, corp_id, force_refresh=False):  # pylint: disable=unused-argument
    corp = CorporationAudit.objects.get(corporation__corporation_id=corp_id)
    logger.debug("Processing Audit Updates for %s", corp.corporation.corporation_name)
    que = []

    que.append(update_corp_wallet.si(corp_id, force_refresh=force_refresh))

    enqueue_next_task(que)

    logger.debug("Queued Audit Updates for %s", corp.corporation.corporation_name)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["corp_id"]},
    name="tasks.update_corp_wallet",
)
@no_fail_chain
def update_corp_wallet(
    self, corp_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_corp_wallet_division(corp_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["character_id"]},
    name="tasks.update_char_mining_ledger",
)
@no_fail_chain
def update_char_mining_ledger(
    self, character_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_character_mining(character_id, force_refresh=force_refresh)


@shared_task(
    bind=True,
    base=QueueOnce,
    once={"graceful": False, "keys": ["character_id"]},
    name="tasks.update_char_wallet",
)
@no_fail_chain
def update_char_wallet(
    self, character_id, force_refresh=False, chain=[]
):  # pylint: disable=unused-argument, dangerous-default-value
    return update_character_wallet(character_id, force_refresh=force_refresh)



from unittest.mock import MagicMock, patch

from ninja import NinjaAPI

from django.test import TestCase
from eveuniverse.models import EveMarketPrice

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import add_character_to_user, create_user_from_evecharacter

from ledger.api.character.ledger import LedgerApiEndpoints
from ledger.models.characteraudit import CharacterMiningLedger
from ledger.tests.test_api._billboardchardata import (
    CharmonthlyMarch,
    CharmonthlyMarchMulti,
    CharNoMining,
    Charyearly,
    noData,
)
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all


class ManageApiLedgerCharEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.char_audit_manager",
                "ledger.char_audit_admin_manager",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user3, _ = create_user_from_evecharacter(
            1003,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerApiEndpoints(api=cls.api)

    def test_get_character_billbboard_api(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/billboard/year/2024/month/3/"

        response = self.client.get(url)

        expected_data = CharmonthlyMarch

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_billbboard_api_multi(self):
        chars = EveCharacter.objects.filter(character_id__in=[1002, 1003])
        for char in chars:
            add_character_to_user(self.user, char)
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/billboard/year/2024/month/3/"

        response = self.client.get(url)

        expected_data = CharmonthlyMarchMulti
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_billbboard_api_cache_no_testing(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/billboard/year/2024/month/3/"

        response = self.client.get(url)
        expected_data = CharmonthlyMarch
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_billbboard_api_no_cache_testing(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/billboard/year/2024/month/3/"

        response = self.client.get(url)
        expected_data = CharmonthlyMarch
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_billbboard_api_no_cache_no_testing(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/billboard/year/2024/month/3/"

        response = self.client.get(url)
        expected_data = CharmonthlyMarch
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_billbboard_api_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/1001/billboard/year/2024/month/3/"
        # when
        response = self.client.get(url)
        # then
        expected_data = CharmonthlyMarch
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_billbboard_api_year(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/account/1001/billboard/year/2024/month/0/"
        expected_data = Charyearly
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_billbboard_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/account/1001/billboard/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)

    def test_get_character_billbboard_api_no_data(self):
        # given
        self.client.force_login(self.user3)
        url = "/ledger/api/account/0/billboard/year/2024/month/3/"
        # when
        response = self.client.get(url)
        # then
        expected_data = noData
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_character_billbboard_api_no_mining(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/billboard/year/2024/month/3/"
        CharacterMiningLedger.objects.all().delete()
        EveMarketPrice.objects.all().delete()
        # when
        response = self.client.get(url)
        # then
        expected_data = CharNoMining
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)



from django.db.models import Q

from ledger.api.managers.billboard_manager import BillboardData, BillboardLedger
from ledger.api.managers.core_manager import LedgerDate, LedgerModels, LedgerTotal
from ledger.hooks import get_extension_logger
from ledger.models.corporationaudit import CorporationWalletJournalEntry

logger = get_extension_logger(__name__)


class CorporationProcess:
    """JournalProcess class to process the journal entries."""

    def __init__(self, year, month, corporations=None):
        self.year = year
        self.month = month
        self.corp = corporations if corporations else []

    # pylint: disable=too-many-locals
    def _process_corporation_chars(self, journal):
        # Create the Dicts for each Character
        corporation_dict = {}
        corporation_total = LedgerTotal()

        for main in journal:
            total_bounty = main.get("total_bounty", 0)
            total_ess = main.get("total_ess", 0)
            character_id = main.get("main_character_id", 0)
            character_name = main.get("main_character_name", "Unknown")

            alts = main.get("alts", [])
            summary_amount = total_bounty + total_ess

            if summary_amount > 0:
                corporation_dict[character_id] = {
                    "main_id": character_id,
                    "main_name": character_name,
                    "alt_names": alts,
                    "total_amount": total_bounty,
                    "total_amount_ess": total_ess,
                }

            totals = {
                "total_amount": total_bounty,
                "total_amount_ess": total_ess,
                "total_amount_all": summary_amount,
            }
            # Summary all
            corporation_total.get_data(totals)

        return corporation_dict, corporation_total

    def generate_ledger(self):
        # Get the Filter Settings
        filter_date = Q(date__year=self.year)
        if not self.month == 0:
            filter_date &= Q(date__month=self.month)

        corporation_journal_year = (
            CorporationWalletJournalEntry.objects.filter(filter_date)
            .select_related(
                "first_party",
                "second_party",
            )
            .annotate_ledger(self.corp)
        )

        # Create the Dicts for each Character
        corporation_dict, corporation_total = self._process_corporation_chars(
            corporation_journal_year
        )

        output = []
        output.append(
            {
                "ratting": sorted(
                    list(corporation_dict.values()), key=lambda x: x["main_name"]
                ),
                "total": corporation_total.to_dict(),
            }
        )

        return output

    def generate_billboard(self, corporations):
        # TODO - Move to new System
        # Get the Filter Settings
        filter_date = Q(date__year=self.year)
        if not self.month == 0:
            filter_date &= Q(date__month=self.month)

        corporation_journal = (
            CorporationWalletJournalEntry.objects.filter(filter_date)
            .select_related(
                "first_party",
                "second_party",
            )
            .annotate_ledger(corporations)
        )

        # Create the Dicts for each Character
        corporation_dict, corporation_total = self._process_corporation_chars(
            corporation_journal
        )

        chars_list = list(corporation_journal.values_list("second_party_id", flat=True))

        # Create Data for Billboard
        date_data = LedgerDate(self.year, self.month)
        data = BillboardData(
            corporation_dict=corporation_dict,
            total_amount=corporation_total.total_amount,
        )
        models = LedgerModels(corporation_journal=corporation_journal)

        # Create the Billboard for the Corporation
        ledger = BillboardLedger(date_data, models, data, corp=True)

        billboard_dict = ledger.billboard_corp_ledger(chars_list)

        output = []
        output.append(
            {
                "billboard": billboard_dict,
            }
        )

        return output



from ninja import NinjaAPI

from django.test import TestCase
from esi.models import Token

from allianceauth.corputils.models import CorpMember, CorpStats
from allianceauth.eveonline.models import EveCorporationInfo
from app_utils.testing import create_user_from_evecharacter

from ledger.api.corporation.template import LedgerTemplateApiEndpoints
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all


class ManageApiTemplateCorpEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
        )
        cls.corp = EveCorporationInfo.objects.get(corporation_id=2001)
        cls.token = Token.objects.get(user=cls.user)
        cls.corpstats = CorpStats.objects.create(corp=cls.corp, token=cls.token)
        cls.corpmember = CorpMember.objects.create(
            character_id="1001", character_name="Gneuten", corpstats=cls.corpstats
        )

        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerTemplateApiEndpoints(api=cls.api)

    def test_get_corporation_ledger_template_api_summary_march(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/0/character/0/ledger/template/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "TAX", status_code=200)
        self.assertContains(response, "ESS", status_code=200)
        self.assertContains(response, "Summary", status_code=200)

    def test_get_corporation_ledger_template_api_summary_year(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/0/character/0/ledger/template/year/2024/month/0/?corp=true"

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "TAX", status_code=200)
        self.assertContains(response, "ESS", status_code=200)
        self.assertContains(response, "Summary", status_code=200)
        self.assertContains(response, "2024", status_code=200)

    def test_get_corporation_ledger_template_api_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/0/character/1001/ledger/template/year/2024/month/3/"
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "TAX", status_code=200)
        self.assertContains(response, "ESS", status_code=200)
        self.assertContains(response, "Gneuten", status_code=200)
        self.assertNotContains(response, "2024", status_code=200)

        # Summary
        self.assertContains(response, "200,000", status_code=200)

    def test_get_corporation_ledger_template_api_single_corp(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2001/character/2001/ledger/template/year/2024/month/3/?corp=True"
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "TAX", status_code=200)
        self.assertContains(response, "ESS", status_code=200)
        self.assertContains(response, "Summary - March", status_code=200)

        # Summary
        self.assertContains(response, "1,400,000", status_code=200)
        self.assertContains(response, "200,000", status_code=200)

    def test_get_corporation_ledger_template_api_year(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2001/character/1001/ledger/template/year/2024/month/0/"

        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "200,000", count=2, status_code=200)
        self.assertContains(response, "TAX", status_code=200)
        self.assertContains(response, "ESS", status_code=200)
        self.assertContains(response, "Gneuten", status_code=200)
        self.assertContains(response, "2024", status_code=200)

    def test_get_corporation_template_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/corporation/1001/character/1001/ledger/template/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)



from django.db import models

from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


class CorpAuditQuerySet(models.QuerySet):
    def visible_to(self, user):
        # superusers get all visible
        if user.is_superuser:
            logger.debug(
                "Returning all corps for superuser %s.",
                user,
            )
            return self

        if user.has_perm("ledger.corp_audit_admin_manager"):
            logger.debug("Returning all corps for Corp Audit Admin %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            queries = [models.Q(corporation__corporation_id=char.corporation_id)]

            if user.has_perm("ledger.corp_audit_manager"):
                queries.append(
                    models.Q(corporation__corporation_id=char.corporation_id)
                )

            logger.debug(
                "%s queries for user %s visible corporations.", len(queries), user
            )

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()


class CorpAuditManagerBase(models.Manager):
    def visible_to(self, user):
        return self.get_queryset().visible_to(user)


CorpAuditManager = CorpAuditManagerBase.from_queryset(CorpAuditQuerySet)



from ninja import NinjaAPI

from django.test import TestCase

from app_utils.testing import create_user_from_evecharacter

from ledger.api.corporation.ledger import LedgerApiEndpoints
from ledger.tests.test_api._billboardcorpdata import noData
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all


class ManageApiLedgerCorpEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.corp_audit_manager",
                "ledger.corp_audit_admin_manager",
            ],
        )

        cls.user2, _ = create_user_from_evecharacter(
            1002,
        )

        cls.user3, _ = create_user_from_evecharacter(
            1003,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerApiEndpoints(api=cls.api)

    def test_get_corporation_billbboard_api_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2002/billboard/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_get_corporation_billbboard_api_year_single(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/2002/billboard/year/2024/month/0/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_get_corporation_billbboard_api(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/0/billboard/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_get_corporation_billbboard_api_year(self):
        self.client.force_login(self.user)
        url = "/ledger/api/corporation/0/billboard/year/2024/month/0/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 200)

    def test_get_corporation_billbboard_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/corporation/2001/billboard/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)

    def test_get_corporation_billbboard_api_no_data(self):
        self.client.force_login(self.user3)
        url = "/ledger/api/corporation/0/billboard/year/2024/month/3/"

        response = self.client.get(url)

        expected_data = noData
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), expected_data)

    def test_get_corporation_billbboard_api_not_found(self):
        self.client.force_login(self.user3)
        url = "/ledger/api/corporation/2001/billboard/year/2024/month/3/"

        response = self.client.get(url)

        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.json(), "Permission Denied")



import math
from collections import defaultdict
from dataclasses import dataclass, field
from decimal import Decimal

from django.db.models import DecimalField, Q, Sum, Value
from django.db.models.functions import (
    Coalesce,
    TruncDay,
    TruncHour,
    TruncMonth,
    TruncYear,
)

from ledger.api.helpers import convert_ess_payout
from ledger.api.managers.core_manager import (
    LedgerData,
    LedgerDate,
    LedgerFilter,
    LedgerModels,
)
from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


@dataclass
class BillboardSum:
    # Income
    sum_amount: list = field(default_factory=lambda: ["Ratting"])
    sum_amount_ess: list = field(default_factory=lambda: ["ESS Payout"])
    sum_amount_misc: list = field(default_factory=lambda: ["Miscellaneous"])
    sum_amount_mining: list = field(default_factory=lambda: ["Mining"])
    # Tick - Ratting
    sum_amount_tick: list = field(default_factory=lambda: ["Tick"])
    # Costs
    sum_amount_costs: list = field(default_factory=lambda: ["Costs"])
    sum_amounts_production_costs: list = field(
        default_factory=lambda: ["Production Costs"]
    )
    sum_amounts_market_costs: list = field(default_factory=lambda: ["Market Costs"])
    # Total
    total_sum: int = None
    total_costs: int = None


@dataclass
class _BillboardDict:
    walletcharts: list = field(default_factory=list)
    charts: list = field(default_factory=list)
    rattingbar: list = field(default_factory=list)
    workflowgauge: list = field(default_factory=list)


@dataclass
class BillboardDict:
    standard: _BillboardDict = field(default_factory=_BillboardDict)
    weekly: _BillboardDict = field(default_factory=_BillboardDict)
    hourly: _BillboardDict = field(default_factory=_BillboardDict)


@dataclass
class BillboardData(LedgerData):
    corporation_dict: dict = field(default_factory=dict)
    total_amount: int = 0


@dataclass
class BillboardBar:
    standard: list = field(default_factory=lambda: ["x"])
    weekly: list = field(default_factory=lambda: ["x"])
    daily: list = field(default_factory=lambda: ["x"])
    hourly: list = field(default_factory=lambda: ["x"])


@dataclass
class BillboardTrunc:
    year: tuple = field(default_factory=lambda: (TruncYear("date"), "%Y-%m"))
    month: tuple = field(default_factory=lambda: (TruncMonth("date"), "%Y-%m"))
    week: tuple = field(default_factory=lambda: (TruncDay("date"), "%Y-%m-%d"))
    day: tuple = field(default_factory=lambda: (TruncDay("date"), "%Y-%m-%d"))
    hour: tuple = field(
        default_factory=lambda: (TruncHour("date"), "%Y-%m-%d %H:00:00")
    )


class BillboardLedger:
    def __init__(
        self, date: LedgerDate, models: LedgerModels, data: BillboardData, corp=False
    ):
        self.is_corp = corp
        self.models = models
        self.date = date
        self.data = data
        self.billboard_dict = BillboardDict()
        self.date_billboard = BillboardBar()

    def annotate_days(self, period, billboard_values, tick=False):
        trunctype, period_format = period
        annotations = {"period": trunctype}
        self.tick = tick

        filters = LedgerFilter(self.chars)

        summary = self.process_billboard(
            annotations, filters, period_format, billboard_values
        )

        return summary

    def process_billboard(self, annotations, filters, period_format, billboard_values):
        self.data_dict = defaultdict(lambda: defaultdict(int))
        summary = BillboardSum()
        if self.models.corp_journal:
            self._process_corp_journal(annotations, period_format)

        if not self.is_corp:
            if self.models.char_journal:
                self._process_char_journal(annotations, filters, period_format)

        for period, data in self.data_dict.items():
            # Main Data
            self.data.total_bounty = data.get("total_bounty", 0)
            self.data.total_ess_payout = data.get("total_ess", 0)
            self.data.total_miscellaneous = data.get("total_miscellaneous", 0)
            self.data.total_mining = data.get("total_mining", 0)
            # Costs
            self.data.total_cost = data.get("total_cost", 0)
            self.data.total_market_cost = data.get("total_market_cost", 0)
            self.data.total_production_cost = data.get("total_production_cost", 0)

            summary.sum_amount.append(int(self.data.total_bounty))
            summary.sum_amount_ess.append(int(self.data.total_ess_payout))

            summary.sum_amount_costs.append(int(self.data.total_cost))
            summary.sum_amounts_market_costs.append(int(self.data.total_market_cost))
            summary.sum_amounts_production_costs.append(
                int(self.data.total_production_cost)
            )

            if not self.is_corp:
                summary.sum_amount_misc.append(int(self.data.total_miscellaneous))
                summary.sum_amount_mining.append(int(self.data.total_mining))
            if self.tick:
                summary.sum_amount_tick.append(int(self.data.total_bounty / 3))

            billboard_values.append(period)

        summary.total_sum = self.calculate_total_sum(summary)
        summary.total_costs = self.calculate_total_costs(summary)
        return summary

    def _process_corp_journal(self, annotations, period_format):
        corp_journal = (
            self.models.corp_journal.annotate(**annotations)
            .values("period")
            .annotate_bounty(self.chars)
            .annotate_ess(self.chars)
            .order_by("period")
        )

        for entry in corp_journal:
            period = entry["period"].strftime(period_format)
            if self.is_corp:
                self.data_dict[period]["total_bounty"] = entry["total_bounty"]
            self.data_dict[period]["total_ess"] = entry["total_ess"]

            if not self.is_corp:
                # Convert the ESS Payout for Character
                self.data_dict[period]["total_ess"] = convert_ess_payout(
                    self.data_dict[period]["total_ess"]
                )

    def _process_char_journal(self, annotations, filters, period_format):
        donations_filter = filters.filter_donation & ~Q(first_party_id__in=self.alts)
        char_journal = (
            self.models.char_journal.annotate(**annotations)
            .values("period")
            .annotate(
                total_bounty=Coalesce(
                    Sum("amount", filter=filters.filter_bounty),
                    Value(0),
                    output_field=DecimalField(),
                ),
                total_miscellaneous=Coalesce(
                    Sum(
                        "amount",
                        filter=filters.filter_all_misc
                        | filters.filter_all_missions
                        | donations_filter,
                    ),
                    Value(0),
                    output_field=DecimalField(),
                ),
                total_cost=Coalesce(
                    Sum(
                        "amount",
                        filter=filters.filter_all_costs,
                    ),
                    Value(0),
                    output_field=DecimalField(),
                ),
                total_market_cost=Coalesce(
                    Sum(
                        "amount",
                        filter=filters.filter_market_cost,
                    ),
                    Value(0),
                    output_field=DecimalField(),
                ),
                total_production_cost=Coalesce(
                    Sum(
                        "amount",
                        filter=filters.filter_production_cost,
                    ),
                    Value(0),
                    output_field=DecimalField(),
                ),
            )
            .order_by("period")
        )

        for entry in char_journal:
            period = entry["period"].strftime(period_format)
            self.data_dict[period]["total_bounty"] = entry["total_bounty"]
            self.data_dict[period]["total_miscellaneous"] = entry["total_miscellaneous"]
            self.data_dict[period]["total_cost"] = entry["total_cost"]
            self.data_dict[period]["total_market_cost"] = entry["total_market_cost"]
            self.data_dict[period]["total_production_cost"] = entry[
                "total_production_cost"
            ]

        if self.models.mining_journal:
            for entry in self.models.mining_journal.values("total", "date"):
                period = entry["date"].strftime(period_format)
                self.data_dict[period]["total_mining"] += (
                    entry["total"] if entry["total"] else 0
                )

    def calculate_total_sum(self, summary: BillboardSum):
        total_sum = sum(
            sum(filter(lambda x: isinstance(x, (int, Decimal)), lst))
            for lst in [
                summary.sum_amount,
                summary.sum_amount_ess,
                summary.sum_amount_misc,
                summary.sum_amount_mining,
            ]
        )
        return total_sum

    def calculate_total_costs(self, summary: BillboardSum):
        total_costs = sum(
            sum(filter(lambda x: isinstance(x, (int, Decimal)), lst))
            for lst in [
                summary.sum_amount_costs,
            ]
        )
        return total_costs

    def calculate_total_production_costs(self, summary: BillboardSum):
        total_production_costs = sum(
            sum(filter(lambda x: isinstance(x, (int, Decimal)), lst))
            for lst in [
                summary.sum_amounts_production_costs,
            ]
        )
        return total_production_costs

    def calculate_total_market_costs(self, summary: BillboardSum):
        total_market_costs = sum(
            sum(filter(lambda x: isinstance(x, (int, Decimal)), lst))
            for lst in [
                summary.sum_amounts_market_costs,
            ]
        )
        return total_market_costs

    def calculate_percentages(self, summary: BillboardSum):
        percentages = [
            (
                (
                    sum(filter(lambda x: isinstance(x, (int, Decimal)), lst))
                    / summary.total_sum
                    * 100
                )
                if summary.total_sum > 0
                else 0
            )
            for lst in [
                summary.sum_amount,
                summary.sum_amount_ess,
                summary.sum_amount_misc,
                summary.sum_amount_mining,
            ]
        ]
        return [math.floor(perc * 10) / 10 for perc in percentages]

    def generate_gauge_data(self, rounded_percentages):
        self.billboard_dict.standard.workflowgauge = (
            [
                ["Ratting", rounded_percentages[0]],
                ["ESS Payout", rounded_percentages[1]],
                ["Miscellaneous", rounded_percentages[2]],
                ["Mining", rounded_percentages[3]],
            ]
            if sum(rounded_percentages)
            else None
        )

    def generate_wallet_ratting_bar(
        self, billboard_dict: _BillboardDict, billboard_values, summary: BillboardSum
    ):
        billboard_dict.rattingbar = (
            (
                [
                    billboard_values,
                    summary.sum_amount,
                    summary.sum_amount_ess,
                    summary.sum_amount_misc,
                    summary.sum_amount_mining,
                    summary.sum_amount_tick,
                ]
            )
            if summary.total_sum
            else None
        )

    # Generate Charts Billboard for Character Ledger
    def generate_wallet_charts_data(self, summary: BillboardSum):
        total_costs = abs(summary.total_costs)
        total_income = abs(summary.total_sum)
        total_market_costs = abs(self.calculate_total_market_costs(summary))
        total_production_costs = abs(self.calculate_total_production_costs(summary))

        misc_cost = total_costs - total_market_costs - total_production_costs

        wallet_chart_data = [
            # Earns
            ["Income", total_income],
            # Costs
            ["Market Cost", total_market_costs],
            ["Production Cost", total_production_costs],
            ["Misc. Cost", misc_cost],
        ]

        self.billboard_dict.standard.walletcharts = (
            wallet_chart_data
            if any(item[1] != 0 for item in wallet_chart_data)
            else None
        )

    # Generate Charts Billboard for Corporation Ledger
    def generate_wallet_corps_data(self):
        if not self.data.corporation_dict:
            self.billboard_dict.standard.charts = None
            return
        others_percentage = 0
        others_name = "Others"
        chart_entries = []

        sorted_entries = sorted(
            self.data.corporation_dict.values(),
            key=lambda x: x["total_amount"],
            reverse=True,
        )

        for i, entry in enumerate(sorted_entries, start=1):
            percentage = (entry["total_amount"] / self.data.total_amount) * 100
            if i <= 10:
                chart_entries.append([entry["main_name"], percentage])
            else:
                others_percentage += percentage

        if len(self.data.corporation_dict) > 10:
            chart_entries.append([others_name, others_percentage])

        self.billboard_dict.standard.charts = chart_entries

    # Create the Billboard Char Ledger
    def billboard_char_ledger(self, chars, alts: list):
        self.chars = [char.character_id for char in chars]
        self.alts = alts
        periods = BillboardTrunc()

        if self.date.month == 0:
            summary = self.annotate_days(periods.month, self.date_billboard.standard)
        else:
            summary = self.annotate_days(
                periods.hour, self.date_billboard.hourly, tick=True
            )
            self.generate_wallet_ratting_bar(
                self.billboard_dict.hourly, self.date_billboard.hourly, summary
            )

            summary = self.annotate_days(periods.day, self.date_billboard.standard)

        # Create Gauge Billboard
        rounded_percentages = self.calculate_percentages(summary)
        self.generate_gauge_data(rounded_percentages)

        # Create Ratting Bar Billboard
        self.generate_wallet_ratting_bar(
            self.billboard_dict.standard, self.date_billboard.standard, summary
        )

        # Create Wallet Charts Billboard
        self.generate_wallet_charts_data(summary)

        return self.billboard_dict

    # Create the Billboard Corp Ledger
    def billboard_corp_ledger(self, chars: list):
        self.chars = chars
        periods = BillboardTrunc()

        if self.date.month == 0:
            summary = self.annotate_days(periods.month, self.date_billboard.standard)
        else:
            summary = self.annotate_days(periods.day, self.date_billboard.standard)

        self.generate_wallet_ratting_bar(
            self.billboard_dict.standard, self.date_billboard.standard, summary
        )

        self.generate_wallet_corps_data()

        return self.billboard_dict



"""Custom exceptions."""

from esi.errors import TokenError


class TokenDoesNotExist(TokenError):
    """A token with a specific scope does not exist for a user."""


class MemberNotActive(Exception):
    pass


class DatabaseError(Exception):
    pass


class ESSError(Exception):
    pass


class CustomError(Exception):
    pass


class LedgerImportError(Exception):
    pass



"""
Etag Helpers
"""

import time

from bravado.exception import HTTPNotModified

from django.core.cache import cache

from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)

MAX_ETAG_LIFE = 60 * 60 * 24 * 7  # 7 Days


class NotModifiedError(Exception):
    pass


def get_etag_key(operation):
    return "etag-" + operation._cache_key()


def get_etag_header(operation):
    return cache.get(get_etag_key(operation), False)


def del_etag_header(operation):
    return cache.delete(get_etag_key(operation), False)


def inject_etag_header(operation):
    etag = get_etag_header(operation)
    logger.debug(
        "ETag: get_etag %s - %s - etag:%s",
        operation.operation.operation_id,
        stringify_params(operation),
        etag,
    )
    if etag:
        operation.future.request.headers["If-None-Match"] = etag


def rem_etag_header(operation):
    logger.debug(
        "ETag: rem_etag %s - %s",
        operation.operation.operation_id,
        stringify_params(operation),
    )
    if "If-None-Match" in operation.future.request.headers:
        del operation.future.request.headers["If-None-Match"]
        return True
    return False


def set_etag_header(operation, headers):
    etag_key = get_etag_key(operation)
    etag = headers.headers.get("ETag", None)
    if etag is not None:
        result = cache.set(etag_key, etag, MAX_ETAG_LIFE)
        logger.debug(
            "ETag: set_etag %s - %s - etag: %s - stored: %s",
            operation.operation.operation_id,
            stringify_params(operation),
            etag,
            result,
        )
        return True
    return False


def stringify_params(operation):
    out = []
    for p, v in operation.future.request.params.items():
        out.append(f"{p}: {v}")
    return ", ".join(out)


def handle_etag_headers(operation, headers, force_refresh, etags_incomplete):
    if (
        get_etag_header(operation) == headers.headers.get("ETag")
        and not force_refresh
        and not etags_incomplete
    ):
        logger.debug("Etag: No modified Data for %s", operation.operation.operation_id)
        raise NotModifiedError()

    if force_refresh:
        logger.debug(
            "ETag: Removing Etag %s F:%s - %s",
            operation.operation.operation_id,
            force_refresh,
            stringify_params(operation),
        )
        del_etag_header(operation)
    else:
        logger.debug(
            "ETag: Saving Etag %s F:%s - %s",
            operation.operation.operation_id,
            force_refresh,
            stringify_params(operation),
        )
        set_etag_header(operation, headers)


def handle_page_results(
    operation, current_page, total_pages, etags_incomplete, force_refresh
):
    results = []
    while current_page <= total_pages:
        operation.future.request.params["page"] = current_page
        try:
            if not etags_incomplete and not force_refresh:
                inject_etag_header(operation)
            else:
                rem_etag_header(operation)

            result, headers = operation.result()
            total_pages = int(headers.headers["X-Pages"])
            handle_etag_headers(operation, headers, force_refresh, etags_incomplete)
            results += result
            current_page += 1

        except (HTTPNotModified, NotModifiedError) as e:
            logger.debug(e)
            if isinstance(e, NotModifiedError):
                total_pages = int(headers.headers["X-Pages"])
            else:
                total_pages = int(e.response.headers["X-Pages"])

            if not etags_incomplete:
                current_page += 1
            else:
                current_page = 1
                results = []
                etags_incomplete = False
    return results, current_page, total_pages


def etag_results(operation, token, force_refresh=False):
    _start_tm = time.perf_counter()
    operation.request_config.also_return_response = True
    if token:
        operation.future.request.headers["Authorization"] = (
            "Bearer " + token.valid_access_token()
        )
    if "page" in operation.operation.params:
        current_page = 1
        total_pages = 1
        etags_incomplete = False
        results, current_page, total_pages = handle_page_results(
            operation, current_page, total_pages, etags_incomplete, force_refresh
        )
    else:
        if not force_refresh:
            inject_etag_header(operation)
        try:
            results, headers = operation.result()
        except HTTPNotModified as e:
            logger.debug("ETag: Not Modified %s", operation.operation.operation_id)
            set_etag_header(operation, e.response)
            raise NotModifiedError() from e
        handle_etag_headers(operation, headers, force_refresh, etags_incomplete=False)
    logger.debug(
        "ESI_TIME: OVERALL %s %s %s",
        time.perf_counter() - _start_tm,
        operation.operation.operation_id,
        stringify_params(operation),
    )
    return results



from django.test import TestCase

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from ledger.models.characteraudit import CharacterAudit, CharacterMiningLedger
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all

MODULE_PATH = "ledger.managers.corpaudit_manager"


class CharAuditQuerySetTest(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

    def test_visible_to_superuser(self):
        # given
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
        )
        self.user.is_superuser = True
        self.user.save()
        # when
        result = CharacterAudit.objects.visible_to(self.user)
        expected_result = CharacterAudit.objects.all()
        # then
        self.assertEqual(list(result), list(expected_result))
        # when
        expected_result = EveCharacter.objects.all()
        result = CharacterAudit.objects.visible_eve_characters(self.user)
        # then
        self.assertEqual(list(result), list(expected_result))

    def test_visible_to_char_audit_admin(self):
        # given
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.char_audit_admin_manager",
            ],
        )
        # when
        expected_result = CharacterAudit.objects.all()
        result = CharacterAudit.objects.visible_to(self.user)
        # then
        self.assertEqual(list(result), list(expected_result))
        # when
        expected_result = EveCharacter.objects.all()
        result = CharacterAudit.objects.visible_eve_characters(self.user)
        # then
        self.assertEqual(list(result), list(expected_result))

    def test_visible_to_char_audit_manager(self):
        # given
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.char_audit_manager",
            ],
        )
        # when
        expected_result = CharacterAudit.objects.filter(
            character__character_id__in=[1001, 1004, 1005, 1006]
        )
        result = CharacterAudit.objects.visible_to(self.user)
        # then
        self.assertEqual(list(result), list(expected_result))
        # when
        expected_result = EveCharacter.objects.filter(
            character_id__in=[
                1001,
                1004,
                1005,
                1006,
                1010,
                1011,
                1012,
                1013,
                1014,
                1015,
                1016,
                1017,
                1018,
                1019,
                1020,
                1021,
                1022,
            ]
        )
        result = CharacterAudit.objects.visible_eve_characters(self.user)
        # then
        self.assertEqual(list(result), list(expected_result))

    def test_visible_to_no_access(self):
        # given
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
        )
        # when
        expected_result = CharacterAudit.objects.filter(character__character_id=1001)
        result = CharacterAudit.objects.visible_to(self.user)
        # then
        self.assertEqual(list(result), list(expected_result))
        # when
        expected_result = EveCharacter.objects.filter(character_id=1001)
        result = CharacterAudit.objects.visible_eve_characters(self.user)
        # then
        self.assertEqual(list(result), list(expected_result))

    def test_visible_to_error(self):
        # given
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
            ],
        )
        self.user.profile.main_character = None
        # when
        expected_result = CharacterAudit.objects.filter(character__character_id=9999)
        result = CharacterAudit.objects.visible_to(self.user)
        # then
        self.assertEqual(list(result), list(expected_result))
        # when
        expected_result = EveCharacter.objects.filter(character_id=99999)
        result = CharacterAudit.objects.visible_eve_characters(self.user)
        # then
        self.assertEqual(list(result), list(expected_result))

    def test_annotate_pricing(self):
        # given
        self.user, self.character_ownership = create_user_from_evecharacter(
            1001,
        )
        # when
        expected_result = CharacterMiningLedger.objects.filter(
            character__character__character_id=1001
        )
        result = CharacterMiningLedger.objects.filter(
            character__character__character_id=1001
        ).annotate_pricing()
        # then
        self.assertEqual(list(result), list(expected_result))



from unittest.mock import patch

from ninja import NinjaAPI

from django.test import TestCase

from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import create_user_from_evecharacter

from ledger.api.character.template import LedgerTemplateApiEndpoints
from ledger.tests.testdata.load_allianceauth import load_allianceauth
from ledger.tests.testdata.load_ledger import load_ledger_all


class ManageApiTemplateCharEndpointsTest(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_ledger_all()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=[
                "ledger.basic_access",
                "ledger.char_audit_manager",
                "ledger.char_audit_admin_manager",
            ],
        )
        cls.user2, _ = create_user_from_evecharacter(
            1002,
            permissions=[
                "ledger.basic_access",
            ],
        )

        cls.api = NinjaAPI()
        cls.manage_api_endpoints = LedgerTemplateApiEndpoints(api=cls.api)

    def test_get_character_ledger_template_api(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/account/0/ledger/template/year/2024/month/3/"
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Ratting", status_code=200)
        self.assertContains(response, "Mining", status_code=200)
        self.assertContains(response, "200,000", status_code=200)

        self.assertContains(response, "ESS", status_code=200)
        self.assertContains(response, "1,133,333", status_code=200)

        self.assertContains(response, "Donations", status_code=200)

    def test_get_character_ledger_template_api_single(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/account/1001/ledger/template/year/2024/month/3/"
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Ratting", status_code=200)
        self.assertContains(response, "Mining", status_code=200)
        self.assertContains(response, "200,000", status_code=200)

        self.assertContains(response, "ESS", status_code=200)
        self.assertContains(response, "1,133,333", status_code=200)

        self.assertContains(response, "Donations", status_code=200)

        # Summary
        self.assertContains(response, "200,000", status_code=200)

    def test_get_character_ledger_template_api_year(self):
        # given
        self.client.force_login(self.user)
        url = "/ledger/api/account/1001/ledger/template/year/2024/month/0/"
        # when
        response = self.client.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        self.assertContains(response, "Ratting", status_code=200)

        self.assertContains(response, "ESS", status_code=200)
        self.assertContains(response, "1,133,333", status_code=200)

        self.assertContains(response, "Mining", status_code=200)
        self.assertContains(response, "Donations", status_code=200)

        # Summary
        self.assertContains(response, "300,000", count=1, status_code=200)

    def test_get_character_ledger_api_no_permission(self):
        self.client.force_login(self.user2)
        url = "/ledger/api/account/1001/ledger/template/year/2024/month/3/"

        response = self.client.get(url)

        self.assertContains(response, "Permission Denied", status_code=200)

    def test_get_character_ledger_api_not_exist(self):
        self.client.force_login(self.user)
        url = "/ledger/api/account/999999/ledger/template/year/2024/month/3/"

        response = self.client.get(url)

        self.assertContains(response, "403 Error", status_code=200)



"""App Configuration"""

# Django
from django.apps import AppConfig

# AA Example App
from ledger import __version__


class LedgerConfig(AppConfig):
    """App Config"""

    default_auto_field = "django.db.models.AutoField"
    author = "Geuthur"
    name = "ledger"
    label = "ledger"
    verbose_name = f"Ledger v{__version__}"



from typing import List

from ninja import NinjaAPI

from django.db.models import Q

from ledger.api import schema
from ledger.api.helpers import get_alts_queryset, get_character
from ledger.hooks import get_extension_logger
from ledger.models import CharacterWalletJournalEntry

logger = get_extension_logger(__name__)


class LedgerJournalApiEndpoints:
    tags = ["CharacerJournal"]

    def __init__(self, api: NinjaAPI):

        @api.get(
            "account/{character_id}/wallet/",
            response={200: List[schema.CharacterWalletEvent], 403: str},
            tags=self.tags,
        )
        def get_character_wallet(
            request, character_id: int, type_refs: str = "", page: int = 1
        ):
            if character_id == 0:
                character_id = request.user.profile.main_character.character_id
            response, main = get_character(request, character_id)

            if not response:
                return 403, "Permission Denied"

            characters = get_alts_queryset(main)

            filters = Q(character__character__in=characters)

            wallet_journal = CharacterWalletJournalEntry.objects.filter(
                filters
            ).select_related("first_party", "second_party")
            output = []

            start_count = (page - 1) * 10000
            end_count = page * 10000

            # pylint: disable=duplicate-code
            if type_refs:
                refs = type_refs.split(",")
                wallet_journal = wallet_journal.filter(ref_type__in=refs)

            wallet_journal = wallet_journal[start_count:end_count]

            for w in wallet_journal:
                output.append(
                    {
                        "character": (w.character.character),
                        "id": w.entry_id,
                        "date": w.date,
                        "first_party": {
                            "id": (w.first_party.eve_id),
                            "name": w.first_party.name,
                            "cat": w.first_party.category,
                        },
                        "second_party": {
                            "id": (w.second_party.eve_id),
                            "name": w.second_party.name,
                            "cat": w.second_party.category,
                        },
                        "ref_type": w.ref_type,
                        "amount": w.amount,
                        "balance": w.balance,
                        "reason": w.reason,
                    }
                )
            return output



from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal

from django.db.models import DecimalField, F, Q, Sum
from django.db.models.functions import Coalesce

from ledger.api.helpers import convert_ess_payout, get_alts_queryset
from ledger.api.managers.core_manager import LedgerFilter
from ledger.hooks import get_extension_logger
from ledger.models.characteraudit import (
    CharacterMiningLedger,
    CharacterWalletJournalEntry,
)
from ledger.models.corporationaudit import CorporationWalletJournalEntry
from ledger.view_helpers.core import calculate_ess_stolen, events_filter

logger = get_extension_logger(__name__)


@dataclass
class TemplateData:
    """TemplateData class to store the data."""

    request: any
    main: any
    year: int
    month: int
    current_date: datetime = None

    def __post_init__(self):
        self.ledger_date = self.current_date
        self.ledger_date = self.ledger_date.replace(year=self.year)
        if self.month != 0:
            self.ledger_date = self.ledger_date.replace(month=self.month)


# pylint: disable=too-many-instance-attributes
@dataclass
class TemplateTotalHour:
    """TemplateTotalHour class to store the hourly data."""

    mission_hour: int = 0
    bounty_hour: int = 0
    ess_hour: int = 0
    mining_hour: int = 0
    stolen_hour: int = 0

    contract_hour: int = 0
    transaction_hour: int = 0
    donation_hour: int = 0
    insurance_hour: int = 0

    contract_cost_hour: int = 0
    production_cost_hour: int = 0
    market_cost_hour: int = 0
    traveling_cost_hour: int = 0
    asset_cost_hour: int = 0
    skill_cost_hour: int = 0
    insurance_cost_hour: int = 0
    planetary_cost_hour: int = 0
    loyality_point_cost_hour: int = 0


# pylint: disable=too-many-instance-attributes
@dataclass
class TemplateTotalDay(TemplateTotalHour):
    """TemplateTotalDay class to store the daily data."""

    mission_day: int = 0
    bounty_day: int = 0
    ess_day: int = 0
    mining_day: int = 0
    stolen_day: int = 0

    contract_day: int = 0
    transaction_day: int = 0
    donation_day: int = 0
    insurance_day: int = 0

    contract_cost_day: int = 0
    production_cost_day: int = 0
    market_cost_day: int = 0
    traveling_cost_day: int = 0
    asset_cost_day: int = 0
    skill_cost_day: int = 0
    insurance_cost_day: int = 0
    planetary_cost_day: int = 0
    loyality_point_cost_day: int = 0


# pylint: disable=too-many-instance-attributes
@dataclass
class TemplateTotalCore(TemplateTotalDay):
    """TemplateTotalCore class to store the core data."""

    mission: int = 0
    bounty: int = 0
    ess: int = 0
    mining: int = 0
    stolen: int = 0

    contract: int = 0
    transaction: int = 0
    donation: int = 0
    insurance: int = 0

    contract_cost: int = 0
    production_cost: int = 0
    market_cost: int = 0
    traveling_cost: int = 0
    asset_cost: int = 0
    skill_cost: int = 0
    insurance_cost: int = 0
    planetary_cost: int = 0
    loyality_point_cost: int = 0


@dataclass
class TemplateTotal(TemplateTotalCore):
    """TemplateTotal class to store the data."""

    def to_dict(self):
        attributes = []
        # PvE
        attributes += ["mission", "bounty", "ess", "mining", "stolen"]
        # Misc
        attributes += ["contract", "transaction", "donation", "insurance"]
        # Costs
        attributes += [
            "contract_cost",
            "production_cost",
            "market_cost",
            "traveling_cost",
            "asset_cost",
            "skill_cost",
            "insurance_cost",
            "planetary_cost",
            "loyality_point_cost",
        ]

        result = {}
        for attr in attributes:
            result[attr] = {
                "total_amount": getattr(self, attr),
                "total_amount_day": getattr(self, f"{attr}_day"),
                "total_amount_hour": getattr(self, f"{attr}_hour"),
            }

        return result


class TemplateProcess:
    """TemplateProcess class to process the data."""

    def __init__(self, chars: list, data: TemplateData, show_year=False):
        self.data = data
        self.chars = chars
        self.show_year = show_year
        self.template_dict = {}

    # pylint: disable=duplicate-code
    def character_template(self):
        """
        Create the character template.
        return: dict
        """
        # Create Char List
        chars = [char.character_id for char in self.chars]

        filters = LedgerFilter(chars)

        filter_date = Q(date__year=self.data.year)
        if not self.data.month == 0:
            filter_date &= Q(date__month=self.data.month)

        # Filter the entries for the current day/month
        character_journal = CharacterWalletJournalEntry.objects.filter(
            filters.filter_partys, filter_date
        ).select_related("first_party", "second_party")

        corporation_journal = (
            CorporationWalletJournalEntry.objects.filter(
                filters.filter_partys, filter_date
            )
            .select_related("first_party", "second_party")
            .order_by("-date")
        )

        # Exclude Events to avoid wrong stats
        corporation_journal = events_filter(corporation_journal)
        mining_journal = (
            CharacterMiningLedger.objects.filter(filters.filter_mining, filter_date)
        ).annotate_pricing()

        self._process_characters(character_journal, corporation_journal, mining_journal)
        return self.template_dict

    def corporation_template(self):
        """
        Create the corporation template.
        return: dict
        """

        chars = [char.character_id for char in self.chars]

        filters = LedgerFilter(chars)

        filter_date = Q(date__year=self.data.year)
        if not self.data.month == 0:
            filter_date &= Q(date__month=self.data.month)

        corporation_journal = (
            CorporationWalletJournalEntry.objects.filter(
                filters.filter_second_party, filter_date
            )
            .select_related("first_party", "second_party")
            .order_by("-date")
        )

        self._process_corporation(corporation_journal)
        return self.template_dict

    # Process the character
    def _process_characters(
        self, character_journal, corporation_journal, mining_journal
    ):
        """Process the characters."""
        alts = get_alts_queryset(self.data.main)

        chars_list = [char.character_id for char in alts]

        total_amounts = TemplateTotal().to_dict()

        models = character_journal, corporation_journal, mining_journal

        amounts = self._process_amounts_char(models, chars_list)

        # Add amounts to total_amounts
        for key, value in amounts.items():
            for sub_key, sub_value in value.items():
                total_amounts[key][sub_key] += sub_value

        self._update_template_dict(self.data.main)
        self._generate_amounts_dict(total_amounts)

    # Process the corporation
    def _process_corporation(self, corporation_journal):
        """Process the corporations."""
        total_amounts = TemplateTotal().to_dict()

        amounts = self._process_amounts_corp(self.chars, corporation_journal)

        # Add amounts to total_amounts
        for key, value in amounts.items():
            for sub_key, sub_value in value.items():
                total_amounts[key][sub_key] += sub_value

        self._update_template_dict(self.data.main)
        self._generate_amounts_dict(total_amounts)

    # Aggregate Journal
    def _aggregate_journal(self, journal):
        result = journal.aggregate(
            total_amount=Coalesce(Sum(F("amount")), 0, output_field=DecimalField()),
            total_amount_day=Coalesce(
                Sum(
                    F("amount"),
                    filter=Q(
                        date__year=self.data.current_date.year,
                        date__month=self.data.current_date.month,
                        date__day=self.data.current_date.day,
                    ),
                ),
                0,
                output_field=DecimalField(),
            ),
            total_amount_hour=Coalesce(
                Sum(
                    F("amount"),
                    filter=Q(
                        date__year=self.data.current_date.year,
                        date__month=self.data.current_date.month,
                        date__day=self.data.current_date.day,
                        date__hour=self.data.current_date.hour,
                    ),
                ),
                0,
                output_field=DecimalField(),
            ),
        )
        if self.data.month == 0:
            result["total_amount_day"] = 0
        return result

    # Update Core Dict
    def _update_template_dict(self, char):
        main_name = char.character_name if not self.show_year else "Summary"
        main_id = char.character_id if not self.show_year else 0
        date = (
            str(self.data.ledger_date.year)
            if self.data.month == 0
            else self.data.ledger_date.strftime("%B")
        )
        self.template_dict.update(
            {
                "main_name": main_name,
                "main_id": main_id,
                "date": date,
            }
        )

    # Add Amounts to Dict
    def _generate_amounts_dict(self, amounts):
        """Generate the amounts dictionary."""
        current_day = 365 if self.data.month == 0 else self.data.ledger_date.day
        # Calculate the total sum
        total_sum = sum(
            amounts[key]["total_amount"] for key in amounts if key != "stolen"
        )

        total_current_day_sum = sum(
            amounts[key]["total_amount_day"] for key in amounts if key != "stolen"
        )

        self.template_dict.update(
            {
                key: {
                    sub_key: value
                    for sub_key, value in {
                        "total_amount": round(amounts[key]["total_amount"], 2),
                        "total_amount_day": (
                            round(amounts[key]["total_amount_day"], 2)
                        ),
                        "total_amount_hour": (
                            round(amounts[key]["total_amount_hour"], 2)
                            if key != "mining"
                            # Bad Fix but it works..
                            else round(amounts[key]["total_amount_day"], 2)
                        ),
                        "average_day": round(
                            amounts[key]["total_amount"] / current_day, 2
                        ),
                        "average_hour": round(
                            (amounts[key]["total_amount"] / current_day) / 24, 2
                        ),
                        "average_tick": (round((amounts[key]["total_amount"]) / 20, 2)),
                        "current_day_tick": (
                            round(amounts[key]["total_amount_day"] / 20, 2)
                        ),
                        "average_day_tick": (
                            round(amounts[key]["total_amount"] / current_day / 20, 2)
                        ),
                        "average_hour_tick": round(
                            (amounts[key]["total_amount"] / current_day) / 24 / 20, 2
                        ),
                    }.items()
                    if value != 0
                }
                for key in amounts
            }
        )
        self.template_dict["summary"] = {
            "total_amount": round(total_sum, 2),
            "total_amount_day": round(total_sum / current_day, 2),
            "total_amount_hour": round((total_sum / current_day) / 24, 2),
            "total_current_day": round(total_current_day_sum, 2),
        }

    # Genereate Amounts for each Char
    def _process_amounts_corp(self, chars, corporation_journal):
        amounts = defaultdict(lambda: defaultdict(Decimal))

        char_ids = [char.character_id for char in chars]

        # Adjust the filters to handle multiple character IDs
        filters = LedgerFilter(char_ids)
        all_filters = filters.get_corp_filters()

        # Calculate the amounts for all characters
        for filter_name, filter_query in all_filters.items():
            aggregated_amounts = self._aggregate_journal(
                corporation_journal.filter(filter_query)
            )
            for sub_key, sub_value in aggregated_amounts.items():
                amounts[filter_name][sub_key] += sub_value

        amounts = calculate_ess_stolen(amounts)

        return amounts

    # Generate Amounts for all Chars
    # pylint: disable=too-many-locals
    def _process_amounts_char(self, models, chars_list):
        amounts = defaultdict(lambda: defaultdict(Decimal))

        # Create the filters for all characters
        filters = LedgerFilter(chars_list)
        all_filters = filters.get_all_filters(chars_list)

        # Set the models
        character_journal, corporation_journal, mining_journal = models

        # Calculate the amounts for all characters
        for filter_name, filter_query in all_filters.items():
            aggregated_amounts = self._aggregate_journal(
                character_journal.filter(filter_query)
            )
            for sub_key, sub_value in aggregated_amounts.items():
                amounts[filter_name][sub_key] += sub_value

        # TODO move to core_manager
        # Add Mining to the amounts
        mining_aggregated = (
            mining_journal.filter(filters.filter_mining)
            .values("total", "date")
            .aggregate(
                total_amount=Coalesce(Sum(F("total")), 0, output_field=DecimalField()),
                total_amount_day=Coalesce(
                    Sum(F("total"), filter=Q(date__day=self.data.current_date.day)),
                    0,
                    output_field=DecimalField(),
                ),
            )
        )
        amounts["mining"]["total_amount"] += mining_aggregated["total_amount"]
        amounts["mining"]["total_amount_day"] += mining_aggregated["total_amount_day"]

        # Add ESS to the amounts
        ess_aggregated = self._aggregate_journal(
            corporation_journal.filter(filters.filter_ess)
        )
        for sub_key, sub_value in ess_aggregated.items():
            amounts["ess"][sub_key] += sub_value

        # Convert ESS Payout for Character Ledger
        amounts["ess"]["total_amount"] = convert_ess_payout(
            amounts["ess"]["total_amount"]
        )

        amounts["ess"]["total_amount_day"] = convert_ess_payout(
            amounts["ess"]["total_amount_day"]
        )
        # Calculate the stolen ESS
        amounts = calculate_ess_stolen(amounts)

        return amounts



"""Generate AllianceAuth test objects from allianceauth.json."""

import json
from datetime import date, datetime
from pathlib import Path

from eveuniverse.models import EveMarketPrice, EveSolarSystem, EveType

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from ledger.models.characteraudit import (
    CharacterAudit,
    CharacterMiningLedger,
    CharacterWalletJournalEntry,
)
from ledger.models.corporationaudit import (
    CorporationAudit,
    CorporationWalletDivision,
    CorporationWalletJournalEntry,
)
from ledger.models.general import EveEntity
from ledger.tests.testdata.load_eveuniverse import load_eveuniverse


def _load_eveentity_data():
    with open(Path(__file__).parent / "eveentity.json", encoding="utf-8") as fp:
        return json.load(fp)


_entities_data = _load_eveentity_data()


def load_ledger_all():
    load_eveentity()
    load_eveuniverse()
    load_char_audit()
    load_corp_audit()
    load_char_mining()
    load_char_journal()
    load_corp_journal()


def load_eveentity():
    EveEntity.objects.all().delete()
    for character_info in _entities_data.get("EveEntity"):
        EveEntity.objects.create(
            eve_id=character_info.get("eve_id"),
            name=character_info.get("name"),
            category=character_info.get("category"),
        )


def load_char_audit():
    CharacterAudit.objects.all().delete()
    CharacterAudit.objects.update_or_create(
        id=1, character=EveCharacter.objects.get(character_id=1001)
    )
    CharacterAudit.objects.update_or_create(
        id=2, character=EveCharacter.objects.get(character_id=1002)
    )
    CharacterAudit.objects.update_or_create(
        id=3, character=EveCharacter.objects.get(character_id=1003)
    )


def load_corp_audit():
    CorporationAudit.objects.all().delete()
    CorporationAudit.objects.update_or_create(
        id=1, corporation=EveCorporationInfo.objects.get(corporation_id=2001)
    )
    CorporationAudit.objects.update_or_create(
        id=2, corporation=EveCorporationInfo.objects.get(corporation_id=2002)
    )
    CorporationAudit.objects.update_or_create(
        id=3, corporation=EveCorporationInfo.objects.get(corporation_id=2003)
    )


def load_char_mining():
    CharacterMiningLedger.objects.all().delete()
    EveMarketPrice.objects.all().delete()
    EveMarketPrice.objects.create(
        eve_type=EveType.objects.get(id=17425),
        adjusted_price=1000,
        average_price=1000,
        updated_at=datetime.now(),
    )
    CharacterMiningLedger.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        id="20240316-17425-1001-30004783",
        date=date(2024, 3, 16),
        type=EveType.objects.get(id=17425),
        system=EveSolarSystem.objects.get(id=30004783),
        quantity=100,
    )
    CharacterMiningLedger.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        id="20240316-17423-1001-30004785",
        date=date(2024, 3, 16),
        type=EveType.objects.get(id=17425),
        system=EveSolarSystem.objects.get(id=30002063),
        quantity=100,
    )
    CharacterMiningLedger.objects.create(
        character=CharacterAudit.objects.get(id=2),
        id="20240316-17423-1002-30004785",
        date=date(2024, 3, 16),
        type=EveType.objects.get(id=17425),
        system=EveSolarSystem.objects.get(id=30002063),
        quantity=100,
    )


def load_char_journal():
    CharacterWalletJournalEntry.objects.all().delete()
    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=100_000,
        balance=0,
        context_id=30004783,
        context_id_type="system_id",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=1,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )
    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=100_000,
        balance=0,
        context_id=30004783,
        context_id_type="system_id",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=2,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )
    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=100_000,
        balance=0,
        context_id=30004783,
        context_id_type="system_id",
        date="2024-01-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=3,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=100_000,
        balance=0,
        context_id=20,
        context_id_type="system_id",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=20,
        reason="",
        ref_type="contract_reward",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    # Player Donation
    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=100_000,
        balance=0,
        context_id=30,
        context_id_type="system_id",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=30,
        reason="Test Transfer",
        ref_type="player_donation",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=-100_000,
        balance=0,
        context_id=31,
        context_id_type="system_id",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1001),
        entry_id=31,
        reason="Test Transfer",
        ref_type="player_donation",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=100_000,
        balance=0,
        context_id=31,
        context_id_type="system_id",
        date="2024-03-19T14:00:01Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1001),
        entry_id=31,
        reason="Test Transfer",
        ref_type="player_donation",
        second_party=EveEntity.objects.get(eve_id=1006),
        tax=0,
        tax_receiver_id=0,
    )

    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=-100_000,
        balance=0,
        context_id=40,
        context_id_type="system_id",
        date="2024-03-19T14:00:03Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=40,
        reason="Test Transfer",
        ref_type="transaction_tax",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=-100_000,
        balance=0,
        context_id=50,
        context_id_type="system_id",
        date="2024-03-19T14:00:07Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=50,
        reason="Test Transfer",
        ref_type="industry_job_tax",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    # Market Escrow
    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=-100_000,
        balance=0,
        context_id=60,
        context_id_type="system_id",
        date="2024-03-19T14:00:10Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=60,
        reason="Test Transfer",
        ref_type="market_escrow",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    # Market Escrow
    CharacterWalletJournalEntry.objects.create(
        character=CharacterAudit.objects.get(character__character_name="Gneuten"),
        amount=-100_000,
        balance=0,
        context_id=60,
        context_id_type="system_id",
        date="2024-03-19T14:00:10Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=60,
        reason="Test Transfer",
        ref_type="insurance",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )


def load_corp_journal():
    CorporationWalletJournalEntry.objects.all().delete()
    CorporationWalletDivision.objects.all().delete()
    CorporationWalletDivision.objects.update_or_create(
        id=1,
        corporation=CorporationAudit.objects.get(id=1),
        balance=100_000,
        division=1,
    )
    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=1,
        reason="",
        ref_type="ess_escrow_transfer",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )
    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=2,
        reason="",
        ref_type="ess_escrow_transfer",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=3,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )
    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=4,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    # Year 2023
    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2023-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=101,
        reason="",
        ref_type="ess_escrow_transfer",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )
    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2023-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=102,
        reason="",
        ref_type="ess_escrow_transfer",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2023-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=201,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )
    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2023-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=202,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1001),
        tax=0,
        tax_receiver_id=0,
    )

    # Many Test Entries
    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=203,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1002),
        tax=0,
        tax_receiver_id=0,
    )

    # Many Test Entries
    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=204,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1010),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=205,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1011),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=206,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1012),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=207,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1013),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=208,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1014),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=209,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1015),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=210,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1016),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=211,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1017),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=212,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1018),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=213,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1019),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=214,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1020),
        tax=0,
        tax_receiver_id=0,
    )

    CorporationWalletJournalEntry.objects.create(
        division=CorporationWalletDivision.objects.get(id=1),
        amount=100_000,
        balance=100_000_000,
        context_id=0,
        context_id_type="division",
        date="2024-03-19T14:00:00Z",
        description="Test",
        first_party=EveEntity.objects.get(eve_id=1000125),
        entry_id=215,
        reason="",
        ref_type="bounty_prizes",
        second_party=EveEntity.objects.get(eve_id=1021),
        tax=0,
        tax_receiver_id=0,
    )



"""
App Settings
"""

import sys

# Django
from app_utils.app_settings import clean_setting

IS_TESTING = sys.argv[1:2] == ["test"]

# Set Test Mode True or False

# Set Naming on Auth Hook
LEDGER_APP_NAME = clean_setting("LEDGER_APP_NAME", "Ledger")

# Caching Key for Caching System
STORAGE_BASE_KEY = "ledger_storage_"

# zKillboard - https://zkillboard.com/
EVE_BASE_URL = "https://esi.evetech.net/"
EVE_API_URL = "https://esi.evetech.net/latest/"
EVE_BASE_URL_REGEX = r"^http[s]?:\/\/esi.evetech\.net\/"

# fuzzwork
FUZZ_BASE_URL = "https://www.fuzzwork.co.uk/"
FUZZ_API_URL = "https://www.fuzzwork.co.uk/api/"
FUZZ_BASE_URL_REGEX = r"^http[s]?:\/\/(www\.)?fuzzwork\.co\.uk\/"

# If True you need to set up the Logger
LEDGER_LOGGER_USE = clean_setting("LEDGER_LOGGER_USE", False)

# Switch between AA-Corp Stats and CorpStats Two APP
LEDGER_CORPSTATS_TWO = clean_setting("LEDGER_CORPSTATS_TWO", False)

# Max Time to set Char Inactive
LEDGER_CHAR_MAX_INACTIVE_DAYS = clean_setting("LEDGER_CHAR_MAX_INACTIVE_DAYS", 3)

# Set the Corporation Tax for Corporation & CharacterLedger Caluclation of ESS Payout
LEDGER_CORP_TAX = clean_setting("LEDGER_CORP_TAX", 15)



"""
Corporation Audit Model
"""

from typing import List

# Django
from django.db import models

from allianceauth.eveonline.models import EveCorporationInfo

from ledger.hooks import get_extension_logger
from ledger.managers.corpaudit_manager import CorpAuditManager
from ledger.managers.corpjournal_manager import CorpWalletManager
from ledger.models.characteraudit import WalletJournalEntry

logger = get_extension_logger(__name__)


class CorporationAudit(models.Model):

    objects = CorpAuditManager()

    corporation = models.OneToOneField(
        EveCorporationInfo,
        on_delete=models.CASCADE,
        related_name="ledger_corporationaudit",
    )

    last_update_wallet = models.DateTimeField(null=True, default=None, blank=True)

    def __str__(self):
        return f"{self.corporation.corporation_name}'s Corporation Data"

    @classmethod
    def get_esi_scopes(cls) -> List[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            # General
            "esi-search.search_structures.v1",
            "esi-universe.read_structures.v1",
            "esi-characters.read_corporation_roles.v1",
            # Mining
            "esi-industry.read_corporation_mining.v1",
            # wallets
            "esi-wallet.read_corporation_wallets.v1",
            "esi-markets.read_corporation_orders.v1",
            "esi-industry.read_corporation_jobs.v1",
            "esi-corporations.read_divisions.v1",
        ]

    class Meta:
        default_permissions = ()
        permissions = (
            ("corp_audit_manager", "Has access to own Corporation"),
            ("corp_audit_admin_manager", "Has access to all Corporations"),
        )


class CorporationWalletDivision(models.Model):
    corporation = models.ForeignKey(
        CorporationAudit,
        on_delete=models.CASCADE,
        related_name="ledger_corporation_division",
    )
    name = models.CharField(max_length=100, null=True, default=None)
    balance = models.DecimalField(max_digits=20, decimal_places=2)
    division = models.IntegerField()

    class Meta:
        default_permissions = ()


class CorporationWalletJournalEntry(WalletJournalEntry):
    division = models.ForeignKey(CorporationWalletDivision, on_delete=models.CASCADE)

    objects = CorpWalletManager()

    def __str__(self):
        return f"Corporation Wallet Journal: {self.first_party.name} '{self.ref_type}' {self.second_party.name}: {self.amount} isk"

    @classmethod
    def get_visible(cls, user):
        corps_vis = CorporationAudit.objects.visible_to(user)
        return cls.objects.filter(division__corporation__in=corps_vis)



"""
Core Helpers
"""

from functools import wraps

# pylint: disable=no-name-in-module
from celery import signature
from celery_once import AlreadyQueued

from ledger.hooks import get_extension_logger

logger = get_extension_logger(__name__)


def enqueue_next_task(chain):
    """
    Queue next task, and attach the rest of the chain to it.
    """
    while len(chain):
        _t = chain.pop(0)
        _t = signature(_t)
        _t.kwargs.update({"chain": chain})
        try:
            _t.apply_async(priority=9)
        except AlreadyQueued:
            # skip this task as it is already in the queue
            logger.debug("Skipping task as its already queued %s", _t)
            continue
        break


def no_fail_chain(func):
    """
    Decorator to chain tasks provided in the chain kwargs regardless of task failures.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        excp = None
        _ret = None
        try:
            _ret = func(*args, **kwargs)
        except Exception as e:  # pylint: disable=broad-exception-caught
            excp = e
        finally:
            _chn = kwargs.get("chain", [])
            enqueue_next_task(_chn)
            if excp:
                raise excp
        return _ret

    return wrapper



"""
Corporation Audit
"""

from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect
from esi.decorators import token_required

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from ledger.models.corporationaudit import CorporationAudit
from ledger.tasks import update_corp


@login_required
@token_required(scopes=CorporationAudit.get_esi_scopes())
@permission_required(["ledger.admin_access"])
def add_corp(request, token) -> HttpResponse:
    char = get_object_or_404(EveCharacter, character_id=token.character_id)
    corp, _ = EveCorporationInfo.objects.get_or_create(
        corporation_id=char.corporation_id,
        defaults={
            "member_count": 0,
            "corporation_ticker": char.corporation_ticker,
            "corporation_name": char.corporation_name,
        },
    )

    CorporationAudit.objects.update_or_create(corporation=corp)
    update_corp.apply_async(
        args=[char.corporation_id], kwargs={"force_refresh": True}, priority=6
    )
    msg = f"{char.corporation_name} successfully added/updated to Ledger"
    messages.info(request, msg)
    return redirect("ledger:ledger_index")
